#include "GridEvolution.h"
#include "ParallelUtility.h"
#include "PlotStuff.h"

int GridEvolution::debug=0;

// ============================================================================================
/// \brief GridEvolution contructor.
// ============================================================================================
GridEvolution::GridEvolution()
{
  maximumNumberOfTimeLevels=4; // keep at most this many time levels
  numberOfTimeLevels=0;
  current=-1;

  accelerationOrderOfAccuracy=1;  // order of accuracy for the acceleration computation
  velocityOrderOfAccuracy=1;      // order of accurcy for the velocity computation

  time.redim(maximumNumberOfTimeLevels);

  specifiedMotion=noSpecifiedMotion;
  
}

// ============================================================================================
/// \brief GridEvolution destructor
// ============================================================================================
GridEvolution::~GridEvolution()
{
}

// ============================================================================================
/// \brief return the number of time levels current saved
// ============================================================================================
int GridEvolution::
getNumberOfTimeLevels() const
{
  return numberOfTimeLevels;
}


// ============================================================================================
/// \brief return the order of accuracy used to compute the acceleration.
// ============================================================================================
int GridEvolution::
getAccelerationOrderOfAccuracy() const
{
  return accelerationOrderOfAccuracy;
}

// ============================================================================================
/// \brief return the order of accuracy used to compute the velocity.
// ============================================================================================
int GridEvolution::
getVelocityOrderOfAccuracy() const
{
  return velocityOrderOfAccuracy;
}


// ============================================================================================
/// \brief set the order of accuracy used to compute the acceleration.
/// \param order (input) : a positive integer
// ============================================================================================
int GridEvolution::
setAccelerationOrderOfAccuracy( int order )
{
  accelerationOrderOfAccuracy=order;
  return 0;
}

// ============================================================================================
/// \brief set the order of accuracy used to compute the velocity.
/// \param order (input) : a positive integer
// ============================================================================================
int GridEvolution::
setVelocityOrderOfAccuracy( int order )
{
  velocityOrderOfAccuracy=order;
  return 0;
}




// ===================================================================================
/// \brief Add a grid to the time history
///
/// \param x (input) : normally x is the vertex array from a MappedGrid
/// \param t (input) : time to associate with this grid.
// ===================================================================================
// add a new grid at time t
int GridEvolution::
addGrid( const realArray & x, real t )
{
  // debug=3;  // ****
  
  if( numberOfTimeLevels!=0 )
  {
    if( t <= time(current) )
    {
      // -- This could be a grid with t<0 for starting a simulation. --
      //    Add it to the start of the list

      // NOTE: We should generalize this to  adding any number of previous grids
      // NOTE: What happens if we want to replace a grid with an improved one ??
         

      if( t <time(current) && time(current)==0.  && numberOfTimeLevels==1 && current==0 )
      {
        gridList.addElement(x,0); // insert into the start of the list

	current = (current +1) % maximumNumberOfTimeLevels;

        time(current)=time(0);
        time(0)=t;
	
        numberOfTimeLevels=min(numberOfTimeLevels+1,maximumNumberOfTimeLevels);

	if( true || debug & 4 )
	  printF("GridEvolution::addGrid:INFO: adding a grid for t<0, t=%9.3e, "
                 "(current=%i, t=%9.3e, numberOfTimeLevels=%i)\n",t,
		 current,time(current),numberOfTimeLevels);
	
        return 1;
      }

      printF("GridEvolution::addGrid:ERROR: t=%10.4e is less than or equal to time(current)=%10.4e \n",
            t,time(current));
      printF("GridEvolution::addGrid:... skipping this grid...\n");
      return 0;
    }
  }

  current = (current +1) % maximumNumberOfTimeLevels;
  
  
  if( numberOfTimeLevels<maximumNumberOfTimeLevels )
  {
    gridList.addElement(x,current);
  }
  else
  { // replace existing element:
   gridList.deleteElement(current);
   gridList.addElement(x,current);
  }
  
  time(current)=t;

  numberOfTimeLevels=min(numberOfTimeLevels+1,maximumNumberOfTimeLevels);
  
  if( debug & 4 )
    printF(" *** GridEvolution::addGrid t=%9.3e, current=%i, numberOfTimeLevels=%i\n",t,current,numberOfTimeLevels);
  
  return 0;
}

// ===========================================================================================
/// \brief Compute the grid velocity from a set of grids over time.
/// \details The GridEvolution class keeps a sequence of past grids and computes the time derivative 
/// of the grid motion using these grids. 
// ==================================================================================================
int GridEvolution::
getVelocity( real t, realSerialArray & gridVelocity, 
	     const Index &I1, const Index &I2, const Index &I3 ) const
{
  if( numberOfTimeLevels>=2 )
  {
    assert( velocityOrderOfAccuracy>0 );

    int i=(current-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
    for( int j=0; j<numberOfTimeLevels-1; j++ )
    {
      if( time(i)<=t )
      {
	break;
      }
      i = (i-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
    }
    
    // we should have time(i) < t < time(i+1) 


    Range Rx = gridVelocity.dimension(3);
    if( velocityOrderOfAccuracy==1 || numberOfTimeLevels==2 )
    {
      if( velocityOrderOfAccuracy != 1 )
	printF("GridEvolution::getVelocity:WARNING: t=%9.3e : only computing the velocity to 1st order accuracy\n"
	       " since there are only 2 time levels available, requested accuracy=%i\n",t,velocityOrderOfAccuracy);

      int ip1 = (i+1) % maximumNumberOfTimeLevels;
      #ifdef USE_PPP
        realSerialArray x0; getLocalArrayWithGhostBoundaries(gridList[i  ],x0);
        realSerialArray x1; getLocalArrayWithGhostBoundaries(gridList[ip1],x1);
      #else
        realSerialArray & x0 = gridList[i  ];
        realSerialArray & x1 = gridList[ip1];
      #endif
      real dt0= time(ip1)-time(i);
      assert( dt0>0. );
      gridVelocity(I1,I2,I3,Rx) = (x1(I1,I2,I3,Rx) - x0(I1,I2,I3,Rx))/dt0;

    }
    else 
    {
      assert( numberOfTimeLevels>=3 );
      if( velocityOrderOfAccuracy != 2 )
	printF("GridEvolution::getVelocity:WARNING: only computing the velocity to 2nd order accuracy"
	       " requested accuracy=%i\n",velocityOrderOfAccuracy);

      int im1 = (i-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
      int ip1 = (i+1) % maximumNumberOfTimeLevels;
      // realArray & x0 = gridList[im1];
      // realArray & x1 = gridList[i];
      // realArray & x2 = gridList[ip1];
      #ifdef USE_PPP
        realSerialArray x0; getLocalArrayWithGhostBoundaries(gridList[im1],x0);
        realSerialArray x1; getLocalArrayWithGhostBoundaries(gridList[i  ],x1);
        realSerialArray x2; getLocalArrayWithGhostBoundaries(gridList[ip1],x2);
      #else
        realSerialArray & x0 = gridList[im1];
        realSerialArray & x1 = gridList[i  ];
        realSerialArray & x2 = gridList[ip1];
      #endif
      const real t0=time(im1), t1=time(i), t2=time(ip1);
      real dt0= t1-t0;
      real dt1= t2-t1;
      assert( dt0>0. && dt1>0. );

      // Compute the time derivative of the Lagrange polynomial: 
      //    x(t) = l0(t)*x0 + l1(t)*x1 + l2(t)*x2 
      // where 
      //   l0 = (t-t1)*(t-t2)/( (t0-t1)*(t0-t2) );
      //   l1 = (t-t2)*(t-t0)/( (t1-t2)*(t1-t0) );
      //   l2 = (t-t0)*(t-t1)/( (t2-t0)*(t2-t1) );
      
      real l0t = (2.*t-(t1+t2))/( (t0-t1)*(t0-t2) );
      real l1t = (2.*t-(t2+t0))/( (t1-t2)*(t1-t0) );
      real l2t = (2.*t-(t0+t1))/( (t2-t0)*(t2-t1) );
      
      gridVelocity(I1,I2,I3,Rx) = l0t*x0(I1,I2,I3,Rx) + l1t*x1(I1,I2,I3,Rx) + l2t*x2(I1,I2,I3,Rx);
      
    }
    
//     printf("GridEvolution::getVelocity: t=%9.3e, current=%i i,ip1=%i,%i, "
// 	   "max(vel)=%8.2e min(vel)=%8.2e\n",t,current,i,ip1,max(gridVelocity(I1,I2,I3,Rx)),
//            min(gridVelocity(I1,I2,I3,Rx)));

  }
  else
  {
    printF("GridEvolution::getVelocity:WARNING: there are only %i time levels (t=%9.3e), setting gridVelocity=0.\n",
	   numberOfTimeLevels,t);

    Range Rx=gridVelocity.dimension(3);
    gridVelocity(I1,I2,I3,Rx)=0.;
  }
  
  // -- piston problem
  if( specifiedMotion==linearMotion )
  {
    // F(t) = - (a/p)*t^p 
    // gt = - a*t^{p-1}
    // gtt = -a*(p-1)*t^{p-2}
    // real ap=1., pp=4.;
    // real ap=1., pp=1.;
    // real ap=-1., pp=1.;
    // real gt = -ap*pow(t,pp-1.);

    // F(t) = a*t^p 
    const real ap=specifiedMotionParameters[0], pp=specifiedMotionParameters[1];
    real gt = ap*pp*pow(t,pp-1.);

    printF(" GridEvolution:velocity: ap=%3.1f pp=%3.1f t=%9.3e,  max=%9.3e, min=%9.3e true=%9.3e\n",ap,pp,
	   t, min(gridVelocity(I1,I2,I3,0)), max(gridVelocity(I1,I2,I3,0)), gt );

    Range all;
    
    gridVelocity(all,all,all,0)=gt; // ***********************************************************************

  }

  return 0;
}

// ========================================================================================
///  \brief Compute the grid acceleration.
// ========================================================================================
int GridEvolution::
getAcceleration( real t, realSerialArray & gridAcceleration, 
		 const Index &I1, const Index &I2, const Index &I3 ) const
{
  if( numberOfTimeLevels>=3 )
  {
    assert( accelerationOrderOfAccuracy>0 );

    int i=(current-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
    for( int j=0; j<numberOfTimeLevels-1; j++ )
    {
      if( time(i)<=t )
      {
	break;
      }
      i = (i-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
    }
    
    // we should have time(i) <= t < time(i+1) 

    int ip1 = (i+1                            ) % maximumNumberOfTimeLevels;
    int im1 = (i-1 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;
    int im2 = (i-2 + maximumNumberOfTimeLevels) % maximumNumberOfTimeLevels;

    real d1=time(ip1)-t;
    real d2=time(i  )-t;
    real d3=time(im1)-t;

    OV_GET_SERIAL_ARRAY_CONST(real,gridList[ip1],x1);
    OV_GET_SERIAL_ARRAY_CONST(real,gridList[i  ],x2);
    OV_GET_SERIAL_ARRAY_CONST(real,gridList[im1],x3);
    
    if( false )
      printF(" GridEvolution: accel: current=%i (ip1,i,im1)=(%i,%i,%i) (tp1,t0,tm1)=(%9.3e,%9.3e,%9.3e) t=%9.3e\n",
	     current,ip1,i,im1,time(ip1),time(i),time(im1),t);


    // Range Rx = gridAcceleration.dimension(3);
    Range Rx = x1.dimension(3); // *wdh* 101021   The gridAcceleration array may be bigger sometimes to also hold gttt
    

    // -- For approximations to the acceleration: see cgDoc/mp/fluidStructure/taylor.maple

    if( accelerationOrderOfAccuracy>=2 && numberOfTimeLevels>=4 )
    {
      // Second order approximation using 4 values      
      OV_GET_SERIAL_ARRAY_CONST(real,gridList[im2],x4);

      real d4=time(im2)-t;

      real d12=d1*d1, d13=d1*d1*d1, d22=d2*d2, d32=d3*d3, d42=d4*d4;

      real c1=-2*(d2+d3+d4)/(-d3*d4*d2+d2*d3*d1-d12*d2+d2*d4*d1+d13+d3*d4*d1-d12*d3-d12*d4);
      real c2=2*(d1+d3+d4)/(d1-d2)/(d3*d4-d3*d2+d22-d4*d2);
      real c3=-2*(d1+d4+d2)/(d1-d3)/(-d4*d2+d3*d2+d3*d4-d32);
      real c4=2*(d1+d3+d2)/(d1-d4)/(-d4*d2+d3*d2+d42-d3*d4);

      gridAcceleration(I1,I2,I3,Rx) = c1*x1(I1,I2,I3,Rx) + c2*x2(I1,I2,I3,Rx) + c3*x3(I1,I2,I3,Rx) + c4*x4(I1,I2,I3,Rx);

      if( accelerationOrderOfAccuracy != 2 )
      {
	printF("GridEvolution::getAcceleration:WARNING: only computing the acceleration to 2nd order accuracy"
	       " requested order =%i\n",accelerationOrderOfAccuracy);
      }


    }
    else
    {
      // First order approximation using 3 values      

      real d12=d1*d1, d22=d2*d2, d32=d3*d3;

      real c1=2/(d2*d3-d2*d1+d12-d3*d1);
      real c2=-2/(-d3*d1+d2*d1-d22+d2*d3);
      real c3=2/(-d3*d1+d2*d1+d32-d2*d3);


      gridAcceleration(I1,I2,I3,Rx) = c1*x1(I1,I2,I3,Rx) + c2*x2(I1,I2,I3,Rx) + c3*x3(I1,I2,I3,Rx);
 
      if( accelerationOrderOfAccuracy > 1  )
      {
	printF("GridEvolution::getAcceleration:WARNING: only computing the acceleration to 1st order accuracy"
	       " requested order =%i. t=%9.3e\n",accelerationOrderOfAccuracy,t);
      }

    }

//     gridAcceleration(I1,I2,I3,Rx) = ( (x2(I1,I2,I3,Rx) - x1(I1,I2,I3,Rx))/dt1
// 				     -(x1(I1,I2,I3,Rx) - x0(I1,I2,I3,Rx))/dt0 )/(.5*(dt0+dt1));
    



    
    if( false )
      printf("GridEvolution::getAcceleration: t=%9.3e, current=%i, im1,i,ip1=%i,%i,%i min(acel)=%8.2e "
	     " max(accel)=%9.2e\n",
	     t, current,im1,i,ip1,min(gridAcceleration(I1,I2,I3,Rx)),max(gridAcceleration(I1,I2,I3,Rx)));
    
  }
  else
  {
    printf("GridEvolution::getAcceleration:WARNING: there are only %i time levels (t=%9.3e). Setting accel. to ZERO.\n",
	   numberOfTimeLevels,t);

    Range Rx=gridAcceleration.dimension(3);
    gridAcceleration(I1,I2,I3,Rx)=0.;
  }
  

  // -- piston problem
  if( specifiedMotion==linearMotion )
  {
    // F(t) = - (a/p)*t^p 
    // gt = - a*t^{p-1}
    // gtt = -a*(p-1)*t^{p-2}
    // real ap=1., pp=4.;
    // real gtt = -ap*(pp-1.)*pow(t,pp-2.);
    // real ap=1., pp=1., gtt=0.;
    // real ap=-1., pp=1., gtt=0.;
    // const real ap=specifiedMotionParameters[0], pp=specifiedMotionParameters[1];
    // real gtt = -ap*(pp-1.)*pow(t,max(pp-2.,0.));

    // F(t) = a*t^p 
    const real ap=specifiedMotionParameters[0], pp=specifiedMotionParameters[1];
    real gtt = ap*pp*(pp-1.)*pow(t,max(pp-2.,0.));


    printF(" GridEvolution:acceleration: ap=%8.2e, pp=%8.2e, t=%9.3e  max=%9.3e, min=%9.3e true=%9.3e\n",
	   ap,pp,t, min(gridAcceleration(I1,I2,I3,0)), max(gridAcceleration(I1,I2,I3,0)), gtt );

    // gridAcceleration(I1,I2,I3,0)=gtt; // ***********************************************************************
    Range all;
    gridAcceleration(all,all,all,0)=gtt; // ***********************************************************************

  }
      

  return 0;
}


// =======================================================================================
/// \brief Interactively update the GridEvolution parameters
// =======================================================================================
int GridEvolution::
update(GenericGraphicsInterface & gi )
{

  // Build a dialog menu for changing parameters
  GUIState gui;
  DialogData & dialog=gui;

  dialog.setWindowTitle("GridEvolution");
  dialog.setExitCommand("exit", "exit");

  dialog.setOptionMenuColumns(1);
  aString specifiedMotionLabel[] = {"no specified motion", "linear motion", "" };
  dialog.addOptionMenu("specified motion:", specifiedMotionLabel, specifiedMotionLabel,(int)specifiedMotion );

//   aString pbLabels[] = {"no specified motion",
//                         "linear motion",
// 			""};
//   int numRows=1;
//   dialog.setPushButtons( pbLabels, pbLabels, numRows ); 

//   aString tbCommands[] = {"save show file",
//                            ""};
//   int tbState[10];
//   tbState[0] = saveShowFile==true;
//   int numColumns=1;
//   dialog.setToggleButtons(tbCommands, tbCommands, tbState, numColumns); 

  // ----- Text strings ------
  const int numberOfTextStrings=5;
  aString textCommands[numberOfTextStrings];
  aString textStrings[numberOfTextStrings];

  int nt=0;
  textCommands[nt] = "velocity order:"; sPrintF(textStrings[nt], "%i",velocityOrderOfAccuracy);  nt++; 
  textCommands[nt] = "acceleration order:"; sPrintF(textStrings[nt], "%i",accelerationOrderOfAccuracy);  nt++; 

  // null strings terminal list
  textCommands[nt]="";   textStrings[nt]="";  assert( nt<numberOfTextStrings );
  dialog.setTextBoxes(textCommands, textCommands, textStrings);

  gi.pushGUI(gui);
  aString answer;
  int len=0;
  for(;;) 
  {
    gi.getAnswer(answer,"");      
    if( answer=="exit" )
    {
      break;
    }
    else if( dialog.getTextValue(answer,"velocity order:","%i",velocityOrderOfAccuracy) )
    {
      printF("Setting the order of accuracy for computing the velocity to %i\n",velocityOrderOfAccuracy);
    }
    else if( dialog.getTextValue(answer,"acceleration order:","%i",accelerationOrderOfAccuracy) )
    {
      printF("Setting the order of accuracy for computing the acceleration to %i\n",accelerationOrderOfAccuracy);
    }
    else if( answer=="no specified motion" )
    {
      printF("Setting specified motion option to `no specified motion'.\n");
    }
    else if( answer=="linear motion" )
    {
      specifiedMotion=linearMotion;

      printF("Setting specified motion option to `linear motion'.\n"
             "Define the velocity and acceleration from a specified motion.\n"
             "The linear motion is defined as F(t) = a*t^p\n");
      real ap=1., pp=1.;
      gi.inputString(answer,"Enter a,p");
      sScanF(answer,"%e %e",&ap,&pp);
      printF("Setting a=%9.3e, p=%9.3e\n",ap,pp);
      specifiedMotionParameters[0]=ap;
      specifiedMotionParameters[1]=pp;
      
    }
    else
    {
      printF("GridEvolution::update: ERROR: unknown answer=%s\n",(const char*)answer);
      gi.stopReadingCommandFile();
    }
    
  }

  gi.popGUI();  // pop dialog
  return 0;
}
