// define BOUNDS_CHECK

#include "RigidBodyMotion.h"
#include "GenericGraphicsInterface.h"


//\begin{>RigidBodyMotionInclude.tex}{\subsection{constructor}} 
RigidBodyMotion::
RigidBodyMotion(int numberOfDimensions_ /* = 3 */)
  : R(0,2)
// =======================================================================================
// /Description:
//    Class for keeping track of the position and orientation of a rigid body moving
// under the influence of forces and torques.
//
// /numberOfDimensions (input):  Indicate whether we are in 2D or 3d. NOTE that this 
//   class basically treats all bodies as 3D (so all input/output is for 3d bodies) but
//   a 2D body will be forced to lie in the x-y plane.
//
// /General Usage:
// \begin{itemize}
//  \item Build a RigidBodyMotion object.
//  \item set the properties, mass and moments of inertia, interactively or using setProperties.
//  \item assign initial conditions (setInitialConditions)
//  \item repeatly call integrate to advance the solution; integrate will return the position
//     of the centre of mass and a rotation matrix that can be used with the MatrixTransform Mapping
//   class, for example, to rotate a Mapping.
// \end{itemize}
// /Background:
// The equations of motion in the standard cartesian reference frame are
// \begin{align*}
//   M {d \vv \over dt} &=  \Fv \\
//   {d  \hv\over dt} &= \sum I_i (\dot{\omega}_i \ev_i+ \omega_i \dot{\ev}_i) = \Gv \\
//   \dot{\ev_i} &= \omegav\times\ev_i \qquad (\ev_i\cdot\ev_i=1, ~~\ev_i\cdot\dot{\ev}_i=0)
// \end{align*}
// or
// \begin{align*}
//   M {d \vv \over dt} &=  \Fv \\
//   I_i \dot{\omega}_i -(I_{i+1}-I_{i+2})\omega_{i+1}\omega_{i+2} &= \Gv\cdot\ev_i \\
//   \dot{\ev_i} &= \omegav\times\ev_i 
// \end{align*}
// where the subscripts on $I$ and $\omega$ are to be taken modulo 3 and 
// where $\ev_i$ are the principal axes of inertia and  $\omega_i$ are the angular velocities about
// the axes of inertia. 
// 
// In this form of the equations we integrate the motion of the principle axes, $\ev_i(t)$ over time along with
// the position of the center of mass.
// The rotation matrix that must be applied to the original body to rotate it to time t
//  is simply $E(t)E^{-1}(0)$ where $E$ is the matrix
// with columns being $\ev_i$,
// \[
//    R(t) = E(t) E^{-1}(0) = \begin{bmatrix} \ev_0(t) & \ev_1(t) & \ev_2(t) \end{bmatrix}
//                           \begin{bmatrix} \ev_0^T(0) \\ \ev_1^T(0) \\ \ev_2^T(0) \end{bmatrix}
// \]
// /numberOfDimensions\_ (input) : 2 or 3 .
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  debug = 0 ;
  
  mass=1.;
  numberOfDimensions=numberOfDimensions_;
  mI.redim(3);
  mI=1.;
  density=-1.; // this means the density is not known
  

  timeSteppingMethod=leapFrogTrapezoidal;
  damping=0.;

  if( false )
  { // try this for light particles:
    timeSteppingMethod=improvedEuler; 
    damping=0.;
  }
  
  numberOfSteps=-1;
  current=-1;
  maximumNumberToSave=5;
  numberSaved=0;

  initialConditionsGiven=0;
  
  positionConstraint=positionHasNoConstraint;
  rotationConstraint=rotationHasNoConstraint;
  constraintValues.redim(15);
  constraintValues=0.;

  time.redim(maximumNumberToSave);  time=0.;
  x0.redim(3);                      x0=0.;
  x.redim(3,maximumNumberToSave);   x=0.;
  v.redim(3,maximumNumberToSave);   v=0.;
  v0.redim(3);                      v0=0.;
  f.redim(3,maximumNumberToSave);   f=0.;
  g.redim(3,maximumNumberToSave);   g=0.;

  mI.redim(3);                      mI=1.;
  w.redim(3,maximumNumberToSave);   w=0.;
  w0.redim(3);                      w0=0.;
  e.redim(3,3,maximumNumberToSave); e=0.;   e(0,0,0)=e(1,1,0)=e(2,2,0)=1.;
 
  e0.redim(3,3);                    e0=0.;  e0(0,0)=e0(1,1)=e0(2,2)=1.;
  
  bodyForce.redim(3);  bodyForce=0.;   // const body force

  // We keep track if the correction steps have converged (N.B. for "light" rigid bodies for e.g.)
  relaxCorrectionSteps=false;
  correctionHasConverged=true;
  maximumRelativeCorrection=0.;


  correctionAbsoluteToleranceForce=1.e-10;
  correctionRelativeToleranceForce=1.e-2;
  correctionRelaxationParameterForce=.2;

  correctionAbsoluteToleranceTorque=1.e-10;
  correctionRelativeToleranceTorque=1.e-2;
  correctionRelaxationParameterTorque=.2;

}

RigidBodyMotion::
~RigidBodyMotion()
{
}

// =================================================================================
/// \brief Return the maximum relative change in the moving grid correction scheme.
///    This is usually only an issue for "light" bodies. 
// =================================================================================
real RigidBodyMotion::
getMaximumRelativeCorrection()
{
  return maximumRelativeCorrection; 
}

// =================================================================================
/// \brief Return true if the correction steps for moving grids have converged.
///    This is usually only an issue for "light" bodies. 
// =================================================================================
bool RigidBodyMotion::
getCorrectionHasConverged()
{
  return correctionHasConverged;
}


//\begin{>>RigidBodyMotionInclude.tex}{\subsection{setProperties}} 
int RigidBodyMotion:: 
setProperties(real totalMass, 
	      const RealArray & momentOfInertia, 
	      const int numberOfDimensions_ /* = 3 */ )
// =======================================================================================
// /Description:
//    Specify properties of the rigid body.
//
// /totalMass (input) : mass of the body
// /momentOfInertia(0:2) (input) : moments of inertial. In 2D only momentsOfInertia(2) is required.
//
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  numberOfDimensions=numberOfDimensions_;
  setMass(totalMass);
  mI(R)=momentOfInertia(R);
  
  return 0;
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{centerOfMassHasBeenInitialized}}
bool RigidBodyMotion::
centerOfMassHasBeenInitialized() const 
// =======================================================================================
// /Description:
//   Return true if the centre of mass has been given initial values.
//
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{ 
  return (initialConditionsGiven % 2) == 1; 
} 

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{massHasBeenInitialized}}
bool RigidBodyMotion::
massHasBeenInitialized() const 
// =======================================================================================
// /Description:
//   Return true if the mass has been given an initial value.
//
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{ 
  return ((initialConditionsGiven/2) % 2) == 1; 
} 

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{setInitialCenterOfMass}} 
int RigidBodyMotion::
setInitialCentreOfMass( const RealArray & xInitial )
// =======================================================================================
// /Description:
//   Supply the inital conditions for integrating the equations of motion. 
// 
// /xInitial(0:2) (input) : inital values for centre of mass.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  if( (initialConditionsGiven % 2)==0 )
    initialConditionsGiven+=1;  // this means the center of mass has been specified.
  
  x0(R)=xInitial(R);
  x(R,0)=x0(R);
  return 0;
}


//\begin{>>RigidBodyMotionInclude.tex}{\subsection{setInitialConditions}} 
int RigidBodyMotion::
setInitialConditions(real t0 /* = 0. */, 
                     const RealArray & xInitial /* = Overture::nullRealArray() */ , 
                     const RealArray & vInitial /* = Overture::nullRealArray() */ , 
                     const RealArray & wInitial /* = Overture::nullRealArray() */ ,
                     const RealArray & axesOfInertia /* = Overture::nullRealArray() */ )
// =======================================================================================
// /Description:
//   Supply the inital conditions for integrating the equations of motion. 
// 
// /t0 (input) : inital time.
// /x0(0:2), vInitial(0:2), wInitial(0:2) (input) : inital values for centre of mass, velocity of the centre of mass
// and the angular velocities.
// /axesOfInertia (input) : (not required in 2d) if this array is dimensioned to be at least 3x3 then it will contain
// the initial orientation of the axes of inertia, axesOfInertia(0:2,i) will be the components of the
//  i'th axes of inertia. This also represents the rotation matrix that transforms the coordinate axes
//  into the axes of inertia.
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  numberOfSteps=0;

  current=-1;
  numberSaved=0;
  time(0)=t0;

  if( xInitial.getLength(0)>=3 )
  {
    if( (initialConditionsGiven % 2)==0 )
      initialConditionsGiven+=1;  // this means the center of mass has been specified.
    x0(R)=xInitial(R);
  }
  else
    x0=0.;

  x(R,0)=x0(R);
  if( vInitial.getLength(0)>=3 ) 
    v0(R)=vInitial(R);
  else
    v0(R)=0.;
  
  v(R,0)=v0(R);
  if( wInitial.getLength(0)>=3 ) 
    w0=wInitial(R);
  else
    w0=0.;
  w(R,0)=w0;
  if( numberOfDimensions==3 && axesOfInertia.getLength(0)>2 && axesOfInertia.getLength(1)>2 )
  {
    e0(R,R)=axesOfInertia(R,R);
    e(R,R,0)=e0;
  }
  else
  {
    // default is the identity matrix
    e0=0.;  e0(0,0)=e0(1,1)=e0(2,2)=1.;
    e=0.;   e(0,0,0)=e(1,1,0)=e(2,2,0)=1.;
  }
  
  
  
  return 0;
}


//\begin{>>RigidBodyMotionInclude.tex}{\subsection{momentumTransfer}} 
int RigidBodyMotion::
momentumTransfer( real t0, RealArray & vNew )
// =======================================================================================
// /Description:
//    Redefine the velocity after a collision.
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  // const int previous = (current-1+maximumNumberToSave) % maximumNumberToSave;
  const int next     = (current+1+maximumNumberToSave) % maximumNumberToSave;

//  if( time(current)==t0 )
  {
    v(R,next)=vNew(R);
    const real dt=t0-time(current);

    x(R,current)=x(R,next)-dt*vNew(R);
    if( debug & 1 )
      printF("RigidBodyMotion::momentumTransfer: t0=%e, time(next)=%e, dt=%8.2e\n",t0,time(current),dt);
  }
  


  return 0;
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{setPositionConstraint}}
int RigidBodyMotion:: 
setPositionConstraint( PositionConstraintEnum constraintType, const RealArray & values )
// ==================================================================================
// /Description:
//    Define a constraint on the position (usually the centre of mass) of the rigid body.
// The position can be contrained to lie in a plane, line along a line or be fixed.
//
// \begin{verbatim}
//   enum PositionConstraintEnum
//   {
//     positionHasNoConstraint=0,
//     positionConstrainedToAPlane,
//     positionConstrainedToALine,
//     positionIsFixed
//   };
// \end{verbatim}
// 
// /constraint (input) : type of constraint (from the enum PositionConstraintEnum)
// /values (input) : For positionConstrainedToAPlane, (v11,v12,v13)=values(0:2) 
//   and (v21,v22,v23)=values (3:5) define the two orthonormal vectors of the constraint plane.
//  For positionConstrainedToALine, (v11,v12,v13)=values(0:2) defines the tangent to the constraining line.
//
// /Notes:
//    More generally we could allow the constraint to apply to a point other than the centre of mass.
// For example, the rigid body could be constrained to rotate about a given point. This option needs
// to be added -- constraintValues(0:2) is allocated to save this other "position"
//
//\end{RigidBodyMotionInclude.tex}  
// ==================================================================================
{
  positionConstraint=constraintType;

  if( positionConstraint==positionConstrainedToAPlane )
  {
    if( constraintValues.getLength(0)<9 )
      constraintValues.resize(9); 

    constraintValues(Range(3,5))=values(Range(0,2));
    constraintValues(Range(6,8))=values(Range(3,5));
  }
  else if( positionConstraint== positionConstrainedToALine )
  {
    if( constraintValues.getLength(0)<6 )
      constraintValues.resize(6); 

    constraintValues.resize(max(9,constraintValues.getLength(0)));
    constraintValues(Range(3,5))=values(Range(0,2));
  }
  
  return 0;
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{setRotationConstraint}}
int RigidBodyMotion:: 
setRotationConstraint( RotationConstraintEnum constraintType, const RealArray & values )
// ==================================================================================
// /Description:
//    Define a constraint on the rotation of the rigid body.
// The rotation can be contrained to lie in a plane, line along a line or be fixed.
//
// \begin{verbatim}
//   enum RotationConstraintEnum
//   {
//     rotationHasNoConstraint=0,
//     rotationConstrainedToAPlane,
//     rotationConstrainedToALine,
//     rotationIsFixed
//   };
// \end{verbatim}
//
// /constraint (input) : type of constraint (from the enum RotationConstraintEnum)
// /values (input) : For rotationConstrainedToAPlane, (v11,v12,v13)=values(0:2) 
//   and (v21,v22,v23)=values (3:5) define the two orthonormal vectors of the constraint plane.
//  For rotationConstrainedToALine, (v11,v12,v13)=values(0:2) defines the tangent to the constraining line.
////
//\end{RigidBodyMotionInclude.tex}  
// ==================================================================================
{
  rotationConstraint=constraintType;

  if( rotationConstraint==rotationConstrainedToAPlane )
  {
    if( constraintValues.getLength(0)<15 )
      constraintValues.resize(15); 

    constraintValues(Range(9,11))=values(Range(0,2));
    constraintValues(Range(12,14))=values(Range(3,5));
  }
  else if( rotationConstraint== rotationConstrainedToALine )
  {
    if( constraintValues.getLength(0)<12 )
      constraintValues.resize(12); 

    constraintValues.resize(max(9,constraintValues.getLength(0)));
    constraintValues(Range(9,11))=values(Range(0,2));
  }

  return 0;
}

int RigidBodyMotion:: 
applyConstraints( const int step, const ConstraintApplicationEnum option )
// ==================================================================================
// /Description:
//    This is a protected routine -- apply a constraint to the forces, torques, positions or rotations.
// The position and rotation constraints are imposed at two stages -- the forces and torques are
// adjusted at one stage (option=applyConstraintsToForces) while the actual positions and 
// rotation angles are adjusted
// at the second stage (option=applyConstraintsToPosition)
// 
// /step : apply constraints at this step.
// /option: option=applyConstraintsToForces : apply constraints to the force and torques
//          option=applyConstraintsToPosition : apply constraints to the positions and angular variables
// 
//\end{RigidBodyMotionInclude.tex}  
// ==================================================================================  
{
  if( positionConstraint==positionHasNoConstraint && 
      rotationConstraint==rotationHasNoConstraint )
  {
    return 0; // no constraints
  }
  
  if( option==applyConstraintsToForces )
  {
    // ****************************************************
    // *** apply constraints to the forcing and torques ***
    // ****************************************************

    if( positionConstraint==positionConstrainedToALine )
    {
      // the force should only have a component along the line
      RealArray vector(3);  vector=constraintValues(Range(3,5));
       
      real dot = sum(f(R,step)*vector(R));  // this assumes the tangent to the line is normalized
      f(R,step) = dot*vector(R);
      
      // adjust the torque -- should be parallel to the line as well
      dot = sum(g(R,step)*vector(R));  // this assumes the tangent to the line is normalized
      g(R,step) = dot*vector(R);
      

    }
    else if( positionConstraint==positionConstrainedToAPlane )
    {
      if( numberOfDimensions==3 )
      {
	RealArray v1(3), v2(3), v3(3); 
        v1=constraintValues(Range(3,5)); v2=constraintValues(Range(6,8));
        // v3 is orthogonal to v1 and b2
        v3(0) = v1(1)*v2(2)-v1(2)*v2(1);
        v3(1) = v1(2)*v2(0)-v1(0)*v2(2);
        v3(2) = v1(0)*v2(1)-v1(1)*v2(0);
	
	// subtract off the component of f in the direction of v3 -- this will project f onto the
        // plane spanned by (v1,v2)
	real dot= sum(f(R,step)*v3(R)); 
	f(R,step) -=  dot*v3(R);
      }
    }
    else if( positionConstraint==positionIsFixed )
    {
      f(R,step)=0.;
    }
    
    if( rotationConstraint==rotationIsFixed )
    {
      g(R,step)=0.;
    }
    else if( rotationConstraint==rotationConstrainedToALine )
    {
      // the force should only have a component along the line
       RealArray vector(3);  vector=constraintValues(Range(9,11));

      real dot = sum(g(R,step)*vector(R));  // this assumes the tangent to the line is normalized
      g(R,step) = dot*vector(R);
    
    }
    else if( rotationConstraint==rotationConstrainedToAPlane )
    {
      if( numberOfDimensions==3 )
      {
	RealArray v1(3), v2(3), v3(3); 
        v1=constraintValues(Range(9,11)); v2=constraintValues(Range(12,14));
        // v3 is orthogonal to v1 and b2
        v3(0) = v1(1)*v2(2)-v1(2)*v2(1);
        v3(1) = v1(2)*v2(0)-v1(0)*v2(2);
        v3(2) = v1(0)*v2(1)-v1(1)*v2(0);
	
	// subtract off the component of g in the direction of v3 -- this will project g onto the
        // plane spanned by (v1,v2)
	real dot= sum(g(R,step)*v3(R)); 
	g(R,step) -=  dot*v3(R);
      }
    }
    
  }
  else if( option==applyConstraintsToPosition )
  {

    // *****************************************
    // *** apply constraints to the position ***
    // *****************************************

    if( positionConstraint==positionConstrainedToALine )
    {
      // the position should lie on a line through the initial xCM
      RealArray vector(3);  vector=constraintValues(Range(3,5));
       
      real dot = sum((x(R,step)-x0(R))*vector(R));  // this assumes the tangent to the line is normalized

      x(R,step) = x0(R) + dot*vector(R);
      
      // printf("RigidBodyMotion:applyConstraint: apply constraint to x at step=%i\n",step);
      
    }
    else if( positionConstraint==positionConstrainedToAPlane )
    {
      if( numberOfDimensions==3 )
      {
	RealArray v1(3), v2(3), v3(3); 
        v1=constraintValues(Range(3,5)); v2=constraintValues(Range(6,8));
        // v3 is orthogonal to v1 and b2
        v3(0) = v1(1)*v2(2)-v1(2)*v2(1);
        v3(1) = v1(2)*v2(0)-v1(0)*v2(2);
        v3(2) = v1(0)*v2(1)-v1(1)*v2(0);
	
	// subtract off the component of x-x0 in the direction of v3 -- this will project x onto the
        // plane x0 + Span(v1,v2)
        //   b = [(x-x0).v3]v3
        //   xNew = x - b
        // => (xNew-x0).v3 = 0
	real dot= sum((x(R,step)-x0(R))*v3(R)); 
	x(R,step) -= dot*v3(R);
      }
    }
    else if( positionConstraint==positionIsFixed )
    {
      x(R,step)=x0(R);
    }
    
    if( rotationConstraint==rotationIsFixed )
    {
      w(R,step)=0.;
    }
    else if( rotationConstraint==rotationConstrainedToALine )
    {
      // the force should only have a component along the line
       RealArray vector(3);  vector=constraintValues(Range(9,11));

      real dot = sum((w(R,step)-w0(R))*vector(R));  // this assumes the tangent to the line is normalized
      w(R,step) = w0(R) + dot*vector(R);
    
    }
    else if( rotationConstraint==rotationConstrainedToAPlane )
    {
      if( numberOfDimensions==3 )
      {
	RealArray v1(3), v2(3), v3(3); 
        v1=constraintValues(Range(9,11)); v2=constraintValues(Range(12,14));
        // v3 is orthogonal to v1 and b2
        v3(0) = v1(1)*v2(2)-v1(2)*v2(1);
        v3(1) = v1(2)*v2(0)-v1(0)*v2(2);
        v3(2) = v1(0)*v2(1)-v1(1)*v2(0);
	
	// subtract off the component of g in the direction of v3 -- this will project g onto the
        // plane spanned by (v1,v2)
	real dot= sum((w(R,step)-w0(R))*v3(R)); 
	w(R,step) -=  dot*v3(R);
      }
    }
  }
  else if( option==applyConstraintsToVelocity )
  {
      Overture::abort("applyConstraintsToVelocity: error"); 
  }
  else
  {
    Overture::abort("error");
  }

  
  return 0;
}


//\begin{>>RigidBodyMotionInclude.tex}{\subsection{integrate}} 
int RigidBodyMotion:: 
integrate(real t0, 
          const RealArray & force, 
	  const RealArray & torque,
	  real t, 
          RealArray & xCM, 
	  RealArray & rotation )
// ==================================================================================
// /Description:
//     Integrate the equations of motion from time t0 to time t using force and moment
// information at time t0.
// 
// /t0 (input) : starting time for the time step.
// /force(0:2) (input) : components of the total force at time t0
// /torque(0:2) (input) : components of the torque (about the center of mass) at time t0.
// /t (input) : end time for the time step.
// /xCM(0:2) (output) : position of the centre of mass at time t
// /rotation(0:2,0:2) (output) : rotation matrix at time t. This matrix will be orthonormal (explicitly enforced
//    by this routine).
//   
//
//\end{RigidBodyMotionInclude.tex}  
// ==================================================================================
{
  if( numberOfSteps<0 )
  {
    printf("RigidBodyMotion::integrate:WARNING:setInitialConditions should be called before integrate\n");
    numberOfSteps=0;
  }
  
  numberOfSteps++;

  if( (current>=0 &&  t0<time(current)) || (numberOfSteps==0 && t0<time(0)) )
  {
    printf("RigidBodyMotion::integrate:ERROR: t0=%6.2e < time(current)=%6.2e integrating backwards in time?\n",t0,
	   numberOfSteps==0  ? time(0) : time(current));
    Overture::abort("error");
  }
  

  current = (current+1) % maximumNumberToSave;
  numberSaved=min(numberSaved+1, maximumNumberToSave);

  const int previous = (current-1+maximumNumberToSave) % maximumNumberToSave;
  const int next     = (current+1) % maximumNumberToSave;


  time(current)=t0;

  const real dt=t-t0;
  time(next)=t;

    

  f(R,current)=force(R)+bodyForce(R);
  g(R,current)=torque(R);

  applyConstraints( current,applyConstraintsToForces );
  
  // guess for forces at next time
  if( numberSaved==0 )
  {
    f(R,next)=f(R,current);   
    g(R,next)=g(R,current);
  }
  else
  {
    real dtOld= numberSaved>1 ? time(current)-time(previous) : 0.;
    real dtRatio;
    if( dtOld!=0. )
      dtRatio=dt/dtOld;
    else
      dtRatio=0.;
    real ab1= 1.+dtRatio; // normally 2
    real ab2= -dtRatio;   // normally -1
    if( timeSteppingMethod==improvedEuler )
    {
      f(R,next)=f(R,current);
      g(R,next)=g(R,current);
    }
    else
    {
      f(R,next)=ab1*f(R,current)+ab2*f(R,previous);
      g(R,next)=ab1*g(R,current)+ab2*g(R,previous);
    }
  
  }
  


  RealArray vDot(R);
  vDot(R)=f(R,current)/mass;

  // NOTE: the v array is not used here in computing x, but is used in the corrector.
  if( numberOfSteps==2 )
  {
    // at t=dt we  recompute v(dt) = v(0)+ ( vDot(0)*dt + vDot(dt)*dt )*.5
    v(R,current)=v(R,previous)+( v(R,current)-v(R,previous) + dt*vDot(R) )*.5;
  }
  if( false )
  {
    printf("RB:integrate: current=%i numberOfSteps=%i v0=(%8.2e,%8.2e) \n",current,numberOfSteps,
	   v0(0),v0(1));
  }
  
  if( numberOfSteps==1 )
  {
    v(R,next)=v0(R)+dt*vDot(R); // this is only first order but is fixed at second step.
    x(R,next)=x(R,current)+dt*( v0(R) + .5*dt*vDot(R) );
  }
  else
  {
    if( timeSteppingMethod==leapFrogTrapezoidal )
    {
      v(R,next)=v(R,previous)+(2.*dt)*vDot(R);  // leap frog
    }
    else if( timeSteppingMethod==improvedEuler )
    {
      v(R,next)=v(R,current)+dt*vDot(R);   // forward euler 
    }
    else
    {
      Overture::abort("error");
    }
    // *wdh* 062306 -- add a damping factor      
    x(R,next)=2.*x(R,current)-x(R,previous)+ (dt*dt)*( vDot(R) -damping*dt*(x(R,current)-x(R,previous)) );
  }
  
  if( numberOfDimensions==2 )
  {
    x(2,next)=0.;
    v(2,next)=0.;
  }

  xCM(R)=x(R,next);

  RealArray wDot(R), eDot(R,R);
  int axis;
  for( axis=0; axis<3; axis++ )
  {
    const int axisp1=(axis+1) % 3;
    const int axisp2=(axis+2) % 3;
    wDot(axis)=( (mI(axisp1)-mI(axisp2))*w(axisp1,current)*w(axisp2,current) +
		 g(0,current)*e(0,axis,current)+
		 g(1,current)*e(1,axis,current)+
		 g(2,current)*e(2,axis,current) )/mI(axis);
  }
  
  if( numberOfSteps==2 )
  {
    // at t=dt we  recompute w(dt) = w(0)+ ( wDot(0)*dt + wDot(dt)*dt )*.5
    w(R,current)=w(R,previous)+( w(R,current)-w(R,previous) + dt*wDot(R) )*.5;
  }

  if( numberOfSteps==1 )
  {
    w(R,next)=w(R,current)+wDot(R)*dt;  // this is only first order for w(dt), but we fix this at step 2
  }
  else
  {
    w(R,next)=w(R,previous)+wDot(R)*(2.*dt); 
  }
  
  // eDot =  w X e
  eDot(0,R)=w(1,current)*e(2,R,current)-w(2,current)*e(1,R,current);
  eDot(1,R)=w(2,current)*e(0,R,current)-w(0,current)*e(2,R,current);
  eDot(2,R)=w(0,current)*e(1,R,current)-w(1,current)*e(0,R,current);
  

  if( numberOfSteps==1 )
  {
    RealArray eDotDot(3,3);
    
    eDotDot(0,R)=(wDot(1,current)*   e(2,R,current)-wDot(2,current)*   e(1,R,current)+
		     w(1,current)*eDot(2,R,current)-   w(2,current)*eDot(1,R,current));
    eDotDot(1,R)=(wDot(2,current)*   e(0,R,current)-wDot(0,current)*   e(2,R,current)+
		     w(2,current)*eDot(0,R,current)-   w(0,current)*eDot(2,R,current));
    eDotDot(2,R)=(wDot(0,current)*   e(1,R,current)-wDot(1,current)*   e(0,R,current)+
		     w(0,current)*eDot(1,R,current)-   w(1,current)*eDot(0,R,current));
    
    e(R,R,next)=e(R,R,current)+dt*(eDot(R,R)+.5*dt*eDotDot(R,R));
  }
  else
  {
    e(R,R,next)=e(R,R,previous)+eDot(R,R)*(2.*dt);
  }
  
  if( numberOfDimensions==2 )
  {
    // 2D: first 2 axes should remain in the plane.
    e(2,R,next)=0.;
    e(R,2,next)=0.;
    e(2,2,next)=1.;
  }
  
  // The matrix e(R,R,next) should be orthonormal
  // Renormalize and orthogonalize the axes of inertia
  int n;
  for( n=0; n<3; n++ )
  {
    // make column "n" orthogonal to all previous columns "m"
    for( int m=0; m<n; m++ )
    {
      real dot=e(0,m,next)*e(0,n,next)+e(1,m,next)*e(1,n,next)+e(2,m,next)*e(2,n,next);
      e(R,n,next)-=dot*e(R,m,next);
    }
    // normalized column n
    real normInverse=1./max(REAL_MIN,SQRT( SQR(e(0,n,next))+SQR(e(1,n,next))+SQR(e(2,n,next))));
    e(R,n,next)*=normInverse;
      
  }
  

  for( n=0; n<3; n++ )
    rotation(R,n)=e(R,0,next)*e0(0,n)+e(R,1,next)*e0(1,n)+e(R,2,next)*e0(2,n);

  applyConstraints( next,applyConstraintsToPosition );


  if( false )
  {
    printF("integrate: t=%e, current=%i x(next)=(%6.2e,%6.2e,%6.2e), w(current)=(%6.2e,%6.2e,%6.2e)\n",
	   t,current,x(0,next),x(1,next),x(2,next),w(0,current),w(1,current),w(2,current));
  }
  if( debug & 1 )
  {
    printF("RB: integrate: t0=%e, t=%e, current=%i x(next)=(%6.2e,%6.2e,%6.2e), force=(%6.2e,%6.2e,%6.2e)\n",
	   t0,t,current,x(0,next),x(1,next),x(2,next),force(0),force(1),force(2));
  }
  
  
  return 0;
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{integrate}} 
int RigidBodyMotion:: 
correct(real t, 
	const RealArray & force, 
	const RealArray & torque,
	RealArray & xCM, 
	RealArray & rotation )
// ==================================================================================
// /Description:
//    Correct the solution at time t using new values for the forces and torques.
//  This routine might be called by a predictor-corrector type algorithm.
// 
// /t (input) : time (corresponding to the end of the time step).
// /force(0:2) (input) : components of the total force at time t
// /torque(0:2) (input) : components of the torque (about the center of mass) at time t
// /xCM(0:2) (output) : new positions of the centre of mass at time t
// /rotation(0:2,0:2) (output) : rotation matrix at time t. This matrix will be orthonormal (explicitly enforced
//    by this routine).
//   
//
//\end{RigidBodyMotionInclude.tex}  
// ==================================================================================
{
  correctionHasConverged=true;
  maximumRelativeCorrection=0.;

  if( numberOfSteps<0 )
  {
    printF("RigidBodyMotion::integrate:WARNING:setInitialConditions should be called before integrate\n");
    numberOfSteps=0;
  }
  else if( numberOfSteps==0 )
  {
    // assign forces for t==0
    f(R,0)=force(R)+bodyForce(R);
    g(R,0)=torque(R);
    xCM(R)=x(R,0);
    rotation=0.;
    rotation(0,0)=rotation(1,1)=rotation(2,2)=1.;

    applyConstraints( 0,applyConstraintsToForces );
    return 0;
  }
  
  assert( current>=0 );
  const int next     = (current+1) % maximumNumberToSave;

  if( time(next)!=t )
  {
    printF("RigidBodyMotion::correct:ERROR: t=%6.2e != time(next)=%6.2e\n",t,time(next));
    OV_ABORT("error");
  }
  

  if( timeSteppingMethod==improvedEuler )
  {
    
    // average force with the predicted value *wdh* 062306
    f(R,next)=.5*(force(R)+bodyForce(R) + f(R,next));
    // f(R,next)=.5*(force(R)+bodyForce(R) + f(R,current));
    g(R,next)=torque(R);

  }
  else
  {
    // -- Leap-frog trapezoidal --
 
    if( relaxCorrectionSteps )
    {
      
      // --- under-relax the new value for f(t+dt), force(R), with the old value for f(t+dt), f(R,next). ---
      //  -- This seems to work for "light" bodies.
      

      real fNorm = sqrt(sum(f(R,next)*f(R,next)));
      real fDiff = max(fabs(force(R)-f(R,next)));
      real fDiffRelative = fDiff/max(1.e-5,fNorm);  // ****** note 1.e-5 -- FIX ME ---

      real gNorm = sqrt(sum(g(R,next)*g(R,next)));
      real gDiff = max(fabs(torque(R)-g(R,next)));
      real gDiffRelative = gDiff/max(1.e-5,gNorm);  // ****** note 1.e-5 -- FIX ME ---


      real relativeDiff = max(fDiffRelative,gDiffRelative);
      real absoluteDiff = max(fDiff,gDiff);

      maximumRelativeCorrection=max(maximumRelativeCorrection,relativeDiff );  // fix me -- use relative ? 

      // The corrections are assumed to have converged if the relative OR the absolute tolerance has been met:
      bool forceHasConverged = fDiffRelative<correctionRelativeToleranceForce ||
                               fDiff < correctionAbsoluteToleranceForce;

      bool torqueHasConverged = gDiffRelative<correctionRelativeToleranceTorque ||
                                gDiff < correctionAbsoluteToleranceTorque;

      correctionHasConverged= forceHasConverged && torqueHasConverged;

      const real alpha=correctionRelaxationParameterForce; 
      const real beta =correctionRelaxationParameterTorque; 
      if( debug & 2 )
      {
	printF("RB:correct: relax force : t=%8.2e alpha=%8.2e: fDiff=%8.2e <? %8.2e or fDiff/fNorm=%8.2e <? %8.2e\n"
	       "          : relax torque: t=%8.2e beta =%8.2e: gDiff=%8.2e <? %8.2e or gDiff/gNorm=%8.2e <? %8.2e\n"
	       "           f(old)=(%8.2e,%8.2e,%8.2e) f(new)=(%8.2e,%8.2e,%8.2e), g(new)=(%8.2e,%8.2e,%8.2e)\n",
	       t,alpha,fDiff,correctionAbsoluteToleranceForce,fDiffRelative,correctionRelativeToleranceForce,
               t,beta, gDiff,correctionAbsoluteToleranceTorque,gDiffRelative,correctionRelativeToleranceTorque,
               f(0,next),f(1,next),f(2,next),force(0),force(1),force(2),
	       torque(0),torque(1),torque(2));
      }
      
      f(R,next)=alpha*( force(R)+bodyForce(R) ) + (1.-alpha)*f(R,next);
    
      g(R,next)=beta*torque(R) + (1.-beta)*g(R,next);  

    }
    else
    {
      // no force relaxation: just update the force at the new time with the given values
      f(R,next)=force(R)+bodyForce(R);
      g(R,next)=torque(R);
    }

  }
  
  applyConstraints( next,applyConstraintsToForces );

  

  assert( current!=next );

  const real dt=t-time(current);


  v(R,next)=v(R,current) + .5*dt*( f(R,current)/mass + f(R,next)/mass );

  x(R,next)=x(R,current) + .5*dt*( v(R,next)+v(R,current) );
  
  xCM(R)=x(R,next);

  RealArray wDot(R), wDotCurrent(R), eDot(R,R), eDotCurrent(R,R);
  int axis;
  for( axis=0; axis<3; axis++ )
  {
    const int axisp1=(axis+1) % 3;
    const int axisp2=(axis+2) % 3;
    wDot(axis)=( (mI(axisp1)-mI(axisp2))*w(axisp1,next)*w(axisp2,next) +
		 g(0,next)*e(0,axis,next)+
		 g(1,next)*e(1,axis,next)+
		 g(2,next)*e(2,axis,next) )/mI(axis);
    
    wDotCurrent(axis)=( (mI(axisp1)-mI(axisp2))*w(axisp1,current)*w(axisp2,current)+
			g(0,current)*e(0,axis,current)+
			g(1,current)*e(1,axis,current)+ 
			g(2,current)*e(2,axis,current) )/mI(axis);
  }
  
  w(R,next)=w(R,current)+.5*dt*( wDot(R)+wDotCurrent(R) );

  // eDot =  w X e
  eDot(0,R)=w(1,next)*e(2,R,next)-w(2,next)*e(1,R,next);
  eDot(1,R)=w(2,next)*e(0,R,next)-w(0,next)*e(2,R,next);
  eDot(2,R)=w(0,next)*e(1,R,next)-w(1,next)*e(0,R,next);
  
  eDotCurrent(0,R)=w(1,current)*e(2,R,current)-w(2,current)*e(1,R,current);
  eDotCurrent(1,R)=w(2,current)*e(0,R,current)-w(0,current)*e(2,R,current);
  eDotCurrent(2,R)=w(0,current)*e(1,R,current)-w(1,current)*e(0,R,current);
  

  e(R,R,next)=e(R,R,current)+.5*dt*(eDot(R,R)+eDotCurrent(R,R));

  
  if( numberOfDimensions==2 )
  {
    // 2D: first 2 axes should remain in the plane.
    e(2,R,next)=0.;
    e(R,2,next)=0.;
    e(2,2,next)=1.;
  }
  
  // The matrix e(R,R,next) should be orthonormal
  // Renormalize and orthogonalize the axes of inertia
  int n;
  for( n=0; n<3; n++ )
  {
    // make column "n" orthogonal to all previous columns "m"
    for( int m=0; m<n; m++ )
    {
      real dot=e(0,m,next)*e(0,n,next)+e(1,m,next)*e(1,n,next)+e(2,m,next)*e(2,n,next);
      e(R,n,next)-=dot*e(R,m,next);
    }
    // normalized column n
    real normInverse=1./max(REAL_MIN,SQRT( SQR(e(0,n,next))+SQR(e(1,n,next))+SQR(e(2,n,next))));
    e(R,n,next)*=normInverse;
      
  }
  

  for( n=0; n<3; n++ )
    rotation(R,n)=e(R,0,next)*e0(0,n)+e(R,1,next)*e0(1,n)+e(R,2,next)*e0(2,n);

  applyConstraints( next,applyConstraintsToPosition );

  if( debug & 1 )
  {
    printF("RB: correct: t=%e, current=%i x(next)=(%6.2e,%6.2e,%6.2e), f(next)=(%6.2e,%6.2e,%6.2e)\n",
	   t,current,x(0,next),x(1,next),x(2,next),f(0,next),f(1,next),f(2,next));
  }

  return 0;
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{integrate}} 
real RigidBodyMotion:: 
getTimeStepEstimate() const
// ==================================================================================
// /Description:
//     Estimate the maximum allowable time step.
//
// /Return value:
//    An estimate of the maximum time step. Return -1. if no estimate is available
//   
//
//\end{RigidBodyMotionInclude.tex}  
// ==================================================================================
{
  if( numberOfSteps<2 )
  {
    printF("RigidBodyMotion::getTimeStepEstimate:ERROR: numberOfSteps<2\n");
    return -1.;
  }

  real dtMax=-1.;

  assert( current >=0 );
  const int previous = (current-1+maximumNumberToSave) % maximumNumberToSave;

  // do this for now -- really need to compute eigenvalues of the Jacobian
  real dx = sqrt(sum(SQR(x(R,current)-x(R,previous))));
  real dv = sqrt(sum(SQR(v(R,current)-v(R,previous))));
  real df = sqrt(sum(SQR(f(R,current)-f(R,previous))));
 

  real fx=0., fv=0.; 
  const real eps = sqrt(REAL_MIN*100.);
   
  if( dx>eps )
    fx = df/dx;
     
  if( dv>eps )
    fv = df/dv;
   

   // estimate lambda*dt where lambda is the maximum modulus of the eigenvalues
  real lambda;
  real des = fv*fv+4.*fx;
  if(  des>0. )
  {
    des=sqrt(des);
    lambda = .5*max( fabs(fv+des), fabs(fv-des) );
  }
  else
  {
    lambda = .5*sqrt(fabs(fx));
  }
   
  const real alpha=1.; // bound on stability region -- choose 1 to be safe (probably=2)
  if( lambda>eps )
    dtMax = (alpha*mass)/lambda; 
  else
    dtMax=-1.;

//    printf("RigidBodyMotion::getTimeStepEstimate: v(current)=(%8.2e,%8.2e,%8.2e) v(prev)=(%8.2e,%8.2e,%8.2e)\n",
//  	 v(0,current),v(1,current),v(2,current), v(0,previous),v(1,previous),v(2,previous));
  
  if( false )
  {
    printF("RigidBodyMotion::getTimeStepEstimate: dx=%8.2e, dv=%8.2e, df=%8.2e, fx=%8.2e, fv=%8.2e, "
           "lambda=%8.2e, dtMax=%8.2e\n", dx,dv,df, fx,fv,lambda,dtMax);
  }
  
  return dtMax;
  
}


  
//\begin{>>RigidBodyMotionInclude.tex}{\subsection{setPosition}} 
int RigidBodyMotion::
getInitialConditions(RealArray & xInitial /* = Overture::nullRealArray() */ , 
                     RealArray & vInitial /* = Overture::nullRealArray() */ , 
                     RealArray & wInitial /* = Overture::nullRealArray() */ ,
                     RealArray & axesOfInertia /* = Overture::nullRealArray() */ ) const
// =======================================================================================
// /Description:
//   Get the current values for the inital conditions.
// 
// /x0(0:2), vInitial(0:2), wInitial(0:2) axesOfInertia(0:2,0:2)  (output) : if any of these arrays is
// dimensioned large enough on input then it is filled in.
//
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  if( xInitial.getLength(0)>=3 )
    xInitial(R)=x0;

  if( vInitial.getLength(0)>=3 ) 
    vInitial(R)=v0;
  
  if( wInitial.getLength(0)>=3 ) 
    wInitial(R)=w0;

  if( numberOfDimensions==3 && axesOfInertia.getLength(0)>2 && axesOfInertia.getLength(1)>2 )
    axesOfInertia(R,R)=e0(R,R);
  
  return 0;
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getMass}} 
real RigidBodyMotion::
getMass() const
// =======================================================================================
// /Description:
//    Return the mass.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  return mass;
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{setMass}} 
void RigidBodyMotion::
setMass(const real totalMass ) 
// =======================================================================================
// /Description:
//    Specify the total mass.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  mass=totalMass;
  if( ((initialConditionsGiven/2)%2)==0 )
    initialConditionsGiven +=2; // this means the mass has been set
}


//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getMomentsOfInertia}} 
RealArray  RigidBodyMotion::
getMomentsOfInertia() const
// =======================================================================================
// /Description:
//    Return the 3 moments of inertia.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  return mI;
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getDensity}} 
real RigidBodyMotion::
getDensity() const
// =======================================================================================
// /Description:
//    Return the density, if known. If not known return a negative value.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  return density;
}



//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getAxesOfInertia}} 
int RigidBodyMotion::
getAxesOfInertia( real t, RealArray & axesOfInertia  ) const
// =======================================================================================
// /Description:
//    Return the axes of inertia at time t.
// /t (input):
// /aCM (output) : axes of inertia at time t.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  if( axesOfInertia.getLength(0)<3 || axesOfInertia.getLength(1)<3 )
    axesOfInertia.redim(3,3);

  return getCoordinates(t,Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),
                        axesOfInertia); 
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getAcceleration}} 
int RigidBodyMotion::
getAcceleration( real t, RealArray & aCM  ) const
// =======================================================================================
// /Description:
//    Return the acceleration of the centre of mass.
// /t (input):
// /aCM (output) : acceleration of the centre of mass.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  if( aCM.getLength(0)<3 )
    aCM.redim(3);

  return getCoordinates(t,Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),aCM);
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getAcceleration}} 
int RigidBodyMotion::
getAngularAcceleration( real t, RealArray & omegaDot  ) const
// =======================================================================================
// /Description:
//    Return the angular acceleration about the centre of mass with respect to the standard
//  coordinate axes (NOT the axes of inertia).
// /t (input):
// /omegaDot (output) : angular acceleration about (x,y,z) axes.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  if( omegaDot.getLength(0)<3 )
    omegaDot.redim(3);

  return getCoordinates(t,Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),omegaDot);
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getPosition}} 
int RigidBodyMotion::
getPosition( real t, RealArray & xCM ) const
// =======================================================================================
// /Description:
//    Return the position of the centre of mass.
// /t (input):
// /xCM (output) :  the position of the centre of mass.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  if( xCM.getLength(0)<3 )
    xCM.redim(3);
  return getCoordinates(t,xCM);
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getVelocity}} 
int RigidBodyMotion::
getVelocity( real t, RealArray & vCM  ) const
// =======================================================================================
// /Description:
//    Return the velocity of the centre of mass.
// /t (input):
// /vCM (output) : velocity of the centre of mass.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  if( vCM.getLength(0)<3 )
    vCM.redim(3);
  return getCoordinates(t,Overture::nullRealArray(),vCM);
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getAngularVelocities}} 
int RigidBodyMotion::
getAngularVelocities( real t, RealArray & omega  ) const
// =======================================================================================
// /Description:
//    Return the angular velocities at time t. 
// /t (input):
// /omega (output) : angular velocities about the axes of inertia. 
//   For a 2d problem only omega(2) will be nonzero.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  if( omega.getLength(0)<3 )
    omega.redim(3);
  return getCoordinates(t,Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),omega );
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getAngularVelocities}} 
int RigidBodyMotion::
getRotationMatrix( real t, 
                   RealArray & r,                     
		   RealArray & rDot /* = Overture::nullRealArray() */,
		   RealArray & rDotDot /* = Overture::nullRealArray()  */) const
// =======================================================================================
// /Description:
//    Return the rotation matrix at time t and optionally the first and second time
// derivative of the rotation matrix.
// /t (input):
// /r (output) : rotation matrix at time t (always 3x3)
// /rDot (output) : if dimensioned on input then on output the 
//     time derivative of the rotation matrix at time t (always 3x3)
// /rDotDot (output) : if dimensioned on input then on output the
//      second time derivative of th rotation matrix at time t (always 3x3)
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  if( r.getLength(0)<3 )
    r.redim(3,3);

  getCoordinates(t,Overture::nullRealArray(),Overture::nullRealArray(),r);

  if( rDot.getLength(0)>=3 && rDot.getLength(1)>=3 )
  {
    // rDot = eDot*e0
    // eDot_i = -omega X e_i 
    RealArray ee(3,3),omega(3),eDot(3,3);
    //                   xCM        vCM             rotation   omega   omegaDot      aCM   axesOfInertia 
    getCoordinates(t,Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),omega,Overture::nullRealArray(),Overture::nullRealArray(),ee);

    // eDot =  w X e
    eDot(0,R)=omega(1)*ee(2,R)-omega(2)*ee(1,R);
    eDot(1,R)=omega(2)*ee(0,R)-omega(0)*ee(2,R);
    eDot(2,R)=omega(0)*ee(1,R)-omega(1)*ee(0,R);

    for( int n=0; n<3; n++ )
      rDot(R,n)=eDot(R,0)*e0(0,n)+eDot(R,1)*e0(1,n)+eDot(R,2)*e0(2,n);

  }
  if( rDotDot.getLength(0)>=3 && rDotDot.getLength(1)>=3 )
  {
    // rDotDot = eDotDot*e0
    // eDotDot_i =  omegaDot X e_i + omega X eDot
    RealArray ee(3,3),omega(3),omegaDot(3),eDot(3,3),eDotDot(3,3);
    //                   xCM        vCM             rotation   omega   omegaDot         aCM   axesOfInertia 
    getCoordinates(t,Overture::nullRealArray(),Overture::nullRealArray(),Overture::nullRealArray(),omega,omegaDot,Overture::nullRealArray(),ee);

    eDot(0,R)=omega(1)*ee(2,R)-omega(2)*ee(1,R);
    eDot(1,R)=omega(2)*ee(0,R)-omega(0)*ee(2,R);
    eDot(2,R)=omega(0)*ee(1,R)-omega(1)*ee(0,R);

    eDotDot(0,R)=omegaDot(1)*ee(2,R)-omegaDot(2)*ee(1,R) + omega(1)*eDot(2,R)-omega(2)*eDot(1,R);
    eDotDot(1,R)=omegaDot(2)*ee(0,R)-omegaDot(0)*ee(2,R) + omega(2)*eDot(0,R)-omega(0)*eDot(2,R);
    eDotDot(2,R)=omegaDot(0)*ee(1,R)-omegaDot(1)*ee(0,R) + omega(0)*eDot(1,R)-omega(1)*eDot(0,R);

    for( int n=0; n<3; n++ )
      rDotDot(R,n)=eDotDot(R,0)*e0(0,n)+eDotDot(R,1)*e0(1,n)+eDotDot(R,2)*e0(2,n);

  }
  

  return 0;
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getPointTransformationMatrix}} 
int RigidBodyMotion::
getPointTransformationMatrix( real t, 
			      RealArray & rDotRt /* = Overture::nullRealArray() */,
			      RealArray & rDotDotRt /* = Overture::nullRealArray()  */) const
// =======================================================================================
// /Description:
//
//    The velocity and acceleration of a point $\pv(t)$ belonging to a rigid body are
// \begin{align*}
//      \dot{\pv} &= \vv(t) + \dot{R} R^T \pv \\
//      \ddot{\pv} &= \av + \ddot{R} R^T \pv    
// \end{align*}
//  since $R^T=R^{-1}$. This routine can compute $\dot{R} R^T$ and $\ddot{R} R^T$.
// 
// Use this routine if you have to compute the velocity and accelerations of a large number of 
// points.
//
// /t (input):
// /rDotRt (output) : if dimensioned on input, on output this will be transformation matrix
//        $\dot{R} R^T$ for the velocity
// /rDotDotRt (output) : if dimensioned on input, on output this will be transformation matrix 
//    $\ddot{R} R^T$  for the acceleration.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{

  RealArray r(3,3),rDot(3,3),rDotDot(3,3);
  getRotationMatrix(t,r,rDot,rDotDot);
  
  r=transpose(r);
  if( rDotRt.getLength(0)>=3 && rDotRt.getLength(1)>=3 )
  {
    for( int n=0; n<3; n++ )
      rDotRt(R,n)=rDot(R,0)*r(0,n)+rDot(R,1)*r(1,n)+rDot(R,2)*r(2,n);
  }
  if( rDotDotRt.getLength(0)>=3 && rDotDotRt.getLength(1)>=3 )
  {
    for( int n=0; n<3; n++ )
      rDotDotRt(R,n)=rDotDot(R,0)*r(0,n)+rDotDot(R,1)*r(1,n)+rDotDot(R,2)*r(2,n);
  }

  return 0;
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getPointVelocity}} 
int RigidBodyMotion::
getPointVelocity( real t, const RealArray & p, RealArray & vp) const
// =======================================================================================
// /Description:
//
//    The velocity of a point $\pv(t)$ belonging to a rigid body is
// \begin{align*}
//      \dot{\pv}(t) &= \vv_\cm(t) + \dot{R} R^T \pv(t) \\
// \end{align*}
// /Note: use {\tt getPointTransformationMatrix} if you need values for many points as this
// routine is not efficient.
// /t (input): evaluate at this time.
// /p (output) : a point at time t.
// /vp (output) : velocity of the point.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  RealArray vCM(3),rDotRt(3,3);
  
  getVelocity(t,vCM);
  getPointTransformationMatrix(t,rDotRt);

  vp(R)=vCM(R)+rDotRt(R,0)*p(0)+rDotRt(R,1)*p(1)+rDotRt(R,2)*p(2);
  
  return 0;
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getPointAcceleration}} 
int RigidBodyMotion::
getPointAcceleration( real t, const RealArray & p, RealArray & ap) const
// =======================================================================================
// /Description:
//
//    The acceleration of a point $\pv(t)$ belonging to a rigid body is
// \begin{align*}
//      \ddot{\pv}(t) &= \av_\cm(t) + \ddot{R} R^T \pv(t) \\
// \end{align*}
// /Note: use {\tt getPointTransformationMatrix} if you need values for many points as this
// routine is not efficient.
//
// /t (input): evaluate at this time.
// /p (output) : a point at time t.
// /ap (output) : acceleration of the point.
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  RealArray aCM(3),rDotDotRt(3,3);
  
  getAcceleration(t,aCM);
  getPointTransformationMatrix(t,rDotDotRt);

  ap(R)=aCM(R)+rDotDotRt(R,0)*p(0)+rDotDotRt(R,1)*p(1)+rDotDotRt(R,2)*p(2);
  return 0;
}


//\begin{>>RigidBodyMotionInclude.tex}{\subsection{getCoordinates}}
int RigidBodyMotion::
getCoordinates( real t, 
		RealArray & xCM      /* = Overture::nullRealArray() */, 
		RealArray & vCM      /* = Overture::nullRealArray() */,
		RealArray & rotation /* = Overture::nullRealArray() */, 
		RealArray & omega    /* = Overture::nullRealArray() */, 
		RealArray & omegaDot /* = Overture::nullRealArray() */, 
		RealArray & aCM      /* = Overture::nullRealArray() */,
                RealArray & axesOfInertia /* = Overture::nullRealArray() */ ) const
// =======================================================================================
// /Description:
//     Return one or more 'coordinates' of the rigid body. This routine will interpolate (or extrapolate)
//   existing values to obtain a value at time t.
//
// /t (input) : return the coordinates at this time
// /xCM (output) : if this array is dimensioned large enough then it will return the 
//       position of the center of mass at time t.
// /vCM (output) : if this array is dimensioned large enough then it will return the 
//       position of the velocity of mass at time t.
// /rotation (output) : (always 3x3) if this array is dimensioned then determine the
//   rotation matrix for rotating the body from it's initial position.
// /omega (output) : if this array is dimensioned large enough then it will return the 
//       angular velocities. This array always
//        has 3 components. For a 2d problem only omega(2) will be nonzero.
// /omegaDot (output) : if this array is dimensioned large enough then it will return the 
//       angular accelerations. This array always
//        has 3 components. For a 2d problem only omega(2) will be nonzero.
// /aCM (output) : acceleration of the centre of mass
// /axesOfInertia (output) : axes of inertia
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  if( numberSaved<0 )
  {
    printF("RigidBodyMotion::integrate:WARNING:setInitialConditions should be called first\n");
    Overture::abort("error");
  }

  // printf(" >>>RigidBodyMotion::getCoordinates: current=%i\n",current);

  //printf("----GETCOORDINATES: mI(1)=%16.8e, mI(2)=%16.8e\n", mI(1), mI(2));
  
  // we just interpolate a value from the closest two times we have

  int i=current;
  int ip1;
  real alpha;
  if( numberSaved>=1 )
  {
    for( int j=0; j<numberSaved; j++ )
    {
      if( t<time(i) )
	i=(i-1+maximumNumberToSave) % maximumNumberToSave;
      else
	break;
    }
    if( t<time(i) )
    {
      // no appropriate time was found..
      i=current;
    }
  
    ip1 = (i+1) % maximumNumberToSave;
    alpha = (time(ip1)-t)/max(REAL_MIN,time(ip1)-time(i));
  }
  else
  {
    if( t!=0. )
    {
      printF("RigidBodyMotion::getCoordinates:ERROR: no information available at t=%e\n",t);
      Overture::abort("error");
    }
    i=0;
    ip1=i;
    alpha=0.;
  }
  
  if( fabs(alpha-.5)>.5 )
  {
    printF("RigidBodyMotion::getPosition: warning, extrapolating value, alpha=%e, t=%e \n",alpha,t);
  }
  if( FALSE )
  {
    printF("RigidBodyMotion::getPosition:alpha=%e, t=%e, i=%i, ip1=%i time(i)=%e, time(ip1)=%e \n",
	   alpha,t,i,ip1,time(i),time(ip1));
  }
  
  if( xCM.getLength(0)>=3 )
    xCM(R)=(1.-alpha)*x(R,ip1)+alpha*x(R,i);

  if( vCM.getLength(0)>=3 )
    vCM(R)=(1.-alpha)*v(R,ip1)+alpha*v(R,i);

  if( rotation.getLength(0)>=3 && rotation.getLength(1)>=3  )
  {
    RealArray r(3,3);
    // rotation matrix = e(t) e^{-1}(0)
    r(R,R)=(1.-alpha)*e(R,R,ip1)+alpha*e(R,R,i);

    int n;
    for( n=0; n<3; n++ )
      rotation(R,n)=r(R,0)*e0(0,n)+r(R,1)*e0(1,n)+r(R,2)*e0(2,n);
  }

  if( axesOfInertia.getLength(0)>=3 && axesOfInertia.getLength(1)>=3  )
    axesOfInertia(R,R)=(1.-alpha)*e(R,R,ip1)+alpha*e(R,R,i);
  
  if( omega.getLength(0)>=3 )
    omega(R)=(1.-alpha)*w(R,ip1)+alpha*w(R,i);

  if( aCM.getLength(0)>=3 )
  {
    aCM(R)=((1.-alpha)*f(R,ip1)+alpha*f(R,i))/mass;
    if( false )
      printF("RigidBodyMotion::getAcceleration: i=%i, ip1=%i, alpha=%7.1e, t=%8.2e "
	   "aCM=(%6.2e,%6.2e,%6.2e) \n",i,ip1,alpha,t,aCM(0),aCM(1),aCM(2));
  }
  if( omegaDot.getLength(0)>=3 )
  {
//      if( numberOfDimensions!=2 )
//      {
//        printf("RigidBodyMotion::getAngularAcceleration: not implemented yet in 3d\n");
//        Overture::abort("error");
//      }
    
    int axis;
    // angular-acceleration : 
    for( axis=0; axis<3; axis++ )
    {
      const int axisp1=(axis+1) % 3;
      const int axisp2=(axis+2) % 3;
      omegaDot(axis)=( 
	(1.-alpha)*( (mI(axisp1)-mI(axisp2))*w(axisp1,ip1)*w(axisp2,ip1)+
		     g(0,ip1)*e(0,axis,ip1)+
		     g(1,ip1)*e(1,axis,ip1)+
		     g(2,ip1)*e(2,axis,ip1))
	+alpha *( (mI(axisp1)-mI(axisp2))*w(axisp1,i)*w(axisp2,i)+
		  g(0,i)*e(0,axis,i)+
		  g(1,i)*e(1,axis,i)+
		  g(2,i)*e(2,axis,i))
	)/mI(axis);
    }
  }
  
  return 0;
}


//\begin{>>RigidBodyMotionInclude.tex}{\subsection{get}} 
int RigidBodyMotion::
get( const GenericDataBase & dir, const aString & name)
// =======================================================================================
// /Description:
//  Read parameters to a data-base file
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  GenericDataBase & subDir = *dir.virtualConstructor();
  dir.find(subDir,name,"RigidBodyMotion");

  aString className;
  subDir.get( className,"className" ); 

  subDir.get(mass,"mass"); 
  subDir.get(density,"density"); 
  subDir.get(numberOfDimensions,"numberOfDimensions"); 
  subDir.get(current,"current"); 
  subDir.get(numberOfSteps,"numberOfSteps"); 
  subDir.get(numberSaved,"numberSaved"); 
  subDir.get(maximumNumberToSave,"maximumNumberToSave"); 
  subDir.get(initialConditionsGiven,"initialConditionsGiven"); 

  subDir.get(mI,"mI"); 
  subDir.get(e,"e"); 
  subDir.get(e0,"e0"); 
  subDir.get(x,"x"); 
  subDir.get(v,"v"); 
  subDir.get(x0,"x0"); 
  subDir.get(v0,"v0"); 
  subDir.get(w0,"w0"); 
  subDir.get(f,"f"); 
  subDir.get(g,"g"); 
  subDir.get(w,"w"); 
  subDir.get(time,"time"); 
  subDir.get(bodyForce,"bodyForce"); 

  int temp;
  subDir.get(temp,"positionConstraint");  positionConstraint=(PositionConstraintEnum)temp;
  subDir.get(temp,"rotationConstraint");  rotationConstraint=(RotationConstraintEnum)temp;
  subDir.get(constraintValues,"constraintValues"); 
  
  subDir.get(relaxCorrectionSteps,"relaxCorrectionSteps");
  // bool correctionHasConverged;
  // real maximumRelativeCorrection;
  subDir.get(correctionAbsoluteToleranceForce,"correctionAbsoluteToleranceForce");
  subDir.get(correctionRelativeToleranceForce,"correctionRelativeToleranceForce");
  subDir.get(correctionRelaxationParameterForce,"correctionRelaxationParameterForce");
  
  subDir.get(correctionAbsoluteToleranceTorque,"correctionAbsoluteToleranceTorque");
  subDir.get(correctionRelativeToleranceTorque,"correctionRelativeToleranceTorque");
  subDir.get(correctionRelaxationParameterTorque,"correctionRelaxationParameterTorque");
  

  // printf(" >>>RigidBodyMotion::get: current=%i\n",current);

  delete &subDir;
  return 0;
}


//\begin{>>RigidBodyMotionInclude.tex}{\subsection{putt}} 
int RigidBodyMotion::
put( GenericDataBase & dir, const aString & name) const
// ======================================================================================
// /Description: 
//    Save parameters to a data-base file
//\end{RigidBodyMotionInclude.tex}  
// ======================================================================================
{  
  GenericDataBase & subDir = *dir.virtualConstructor();      // create a derived data-base object
  dir.create(subDir,name,"RigidBodyMotion");                 // create a sub-directory 

  aString className="RigidBodyMotion";
  subDir.put( className,"className" );

  subDir.put(mass,"mass"); 
  subDir.put(density,"density"); 
  subDir.put(numberOfDimensions,"numberOfDimensions"); 
  subDir.put(current,"current"); 
  subDir.put(numberOfSteps,"numberOfSteps"); 
  subDir.put(numberSaved,"numberSaved"); 
  subDir.put(maximumNumberToSave,"maximumNumberToSave"); 
  subDir.put(initialConditionsGiven,"initialConditionsGiven"); 

  subDir.put(mI,"mI"); 
  subDir.put(e,"e"); 
  subDir.put(e0,"e0"); 
  subDir.put(x,"x"); 
  subDir.put(v,"v"); 
  subDir.put(x0,"x0"); 
  subDir.put(v0,"v0"); 
  subDir.put(w0,"w0"); 
  subDir.put(f,"f"); 
  subDir.put(g,"g"); 
  subDir.put(w,"w"); 
  subDir.put(time,"time"); 
  subDir.put(bodyForce,"bodyForce"); 

  subDir.put(positionConstraint,"positionConstraint");  
  subDir.put(rotationConstraint,"rotationConstraint");  
  subDir.put(constraintValues,"constraintValues"); 

  subDir.put(relaxCorrectionSteps,"relaxCorrectionSteps");
  // bool correctionHasConverged;
  // real maximumRelativeCorrection;
  subDir.put(correctionAbsoluteToleranceForce,"correctionAbsoluteToleranceForce");
  subDir.put(correctionRelativeToleranceForce,"correctionRelativeToleranceForce");
  subDir.put(correctionRelaxationParameterForce,"correctionRelaxationParameterForce");
  
  subDir.put(correctionAbsoluteToleranceTorque,"correctionAbsoluteToleranceTorque");
  subDir.put(correctionRelativeToleranceTorque,"correctionRelativeToleranceTorque");
  subDir.put(correctionRelaxationParameterTorque,"correctionRelaxationParameterTorque");

  delete &subDir;
  return 0;  
}

//\begin{>>RigidBodyMotionInclude.tex}{\subsection{update}} 
int RigidBodyMotion::
update( GenericGraphicsInterface & gi )
// =======================================================================================
// /Description:
// 
//\end{RigidBodyMotionInclude.tex}  
//=========================================================================================
{
  aString answer,answer2;
  char buff[80];


  aString menu[] =
  {
    "!Rigid body parameters",
    "mass",
    "density",
    "moments of inertia",
    "axes of inertia",
    "initial centre of mass",
    "initial velocity",
    "initial angular velocity",
    "body force",
     //   "body torque",
    ">constraints",
    "position has no constraint",
    "position is constrained to a plane",
    "position is constrained to a line",
    "position is fixed",
    "rotation has no constraint",
    "rotation is constrained to a plane",
    "rotation is constrained to a line",
    "rotation is fixed",
    "<done",
    "exit",
    ""
  };
  
  GUIState dialog;
  dialog.setWindowTitle("RigidBodyMotion");
  dialog.setExitCommand("exit", "exit");

//   aString colourBoundaryCommands[] = { "colour by bc",
// 			               "colour by share",
// 			               "" };
//   // dialog.addRadioBox("boundaries:",colourBoundaryCommands, colourBoundaryCommands, 0 );
//   dialog.addOptionMenu("boundaries:",colourBoundaryCommands, colourBoundaryCommands, 0 );

//   aString cmds[] = {"mapping parameters...",
// 		    "show parameters",
//                     "plot",
// 		    "help",
// 		    ""};
//   int numberOfPushButtons=3;  // number of entries in cmds
//   int numRows=numberOfPushButtons; // (numberOfPushButtons+1)/2;
//   dialog.setPushButtons( cmds, cmds, numRows ); 

  aString tbCommands[] = {"relax correction steps",
 			  ""};
  int tbState[10];

  tbState[0] = relaxCorrectionSteps;
  int numColumns=1;
  dialog.setToggleButtons(tbCommands, tbCommands, tbState, numColumns);

  const int numberOfTextStrings=10;  // max number allowed
  aString textLabels[numberOfTextStrings];
  aString textStrings[numberOfTextStrings];


  int nt=0;
  textLabels[nt] = "debug:";  sPrintF(textStrings[nt],"%i",debug);  
  nt++; 

  textLabels[nt] = "force relaxation parameter:";  sPrintF(textStrings[nt],"%g",correctionRelaxationParameterForce);  
  nt++; 
  textLabels[nt] = "force relative tol:";  sPrintF(textStrings[nt],"%g",correctionRelativeToleranceForce); 
  nt++; 
  textLabels[nt] = "force absolute tol:";  sPrintF(textStrings[nt],"%g",correctionAbsoluteToleranceForce); 
  nt++; 

  textLabels[nt] = "torque relaxation parameter:";  sPrintF(textStrings[nt],"%g",correctionRelaxationParameterTorque); 
  nt++; 
  textLabels[nt] = "torque relative tol:";  sPrintF(textStrings[nt],"%g",correctionRelativeToleranceTorque); 
  nt++; 
  textLabels[nt] = "torque absolute tol:";  sPrintF(textStrings[nt],"%g",correctionAbsoluteToleranceTorque); 
  nt++; 


  // null strings terminal list
  textLabels[nt]="";   textStrings[nt]="";  assert( nt<numberOfTextStrings );
  dialog.setTextBoxes(textLabels, textLabels, textStrings);


  // add old popup menu
  dialog.buildPopup(menu);
  gi.pushGUI(dialog);


  gi.appendToTheDefaultPrompt("rigidBody>");
  for( ;; )
  {

    // gi.getMenuItem(menu,answer,"choose an option");
    gi.getAnswer(answer,"");  
 
    if( answer=="exit" || answer=="done" )
    {
      break;
    }
    else if( answer=="mass" )
    {
      gi.inputString(answer2,sPrintF(buff,"Enter the body mass (default=%8.2e)",mass));
      if( answer2!="" )
	sScanF(answer2,"%e",&mass);

      setMass(mass);
      
    }
    else if( answer=="density" )
    {
      gi.inputString(answer2,sPrintF(buff,"Enter the body density, -1. if not known, (default=%8.2e)",density));
      if( answer2!="" )
	sScanF(answer2,"%e",&density);
    }
    else if( answer=="moments of inertia" )
    {
      if( numberOfDimensions==2 )
      {
	gi.inputString(answer2,sPrintF(buff,"Enter the moment of inertia (default=%8.2e)",mI(2)));
	if( answer2!="" )
	  sScanF(answer2,"%e",&mI(2));
      }
      else
      {
	gi.inputString(answer2,sPrintF(buff,"Enter the moments of inertia (default=%8.2e,%8.2e,%8.2e)",
             mI(0),mI(1),mI(2)));
	if( answer2!="" )
	  sScanF(answer2,"%e %e %e",&mI(0),&mI(1),&mI(2));
      }
    }
    else if( answer=="initial centre of mass" || answer=="initial center of mass" )
    {
      if( numberOfDimensions==2 )
      {
	gi.inputString(answer2,sPrintF(buff,"Enter the centre of mass (default=%8.2e,%8.2e)",x0(0),x0(1)));
	if( answer2!="" )
	  sScanF(answer2,"%e %e",&x0(0),&x0(1));
      }
      else
      {
	gi.inputString(answer2,sPrintF(buff,"Enter the centre of mass (default=%8.2e,%8.2e,%8.2e)",
             x0(0),x0(1),x0(2)));
	if( answer2!="" )
	  sScanF(answer2,"%e %e %e",&x0(0),&x0(1),&x0(2));
      }
      setInitialCentreOfMass(x0);
      printF("MovingGrids: >>>>  centerOfMassHasBeenInitialized()=%i \n", centerOfMassHasBeenInitialized()); 
    }
    else if( answer=="initial velocity" )
    {
      if( numberOfDimensions==2 )
      {
	gi.inputString(answer2,sPrintF(buff,"Enter the initial velocity (default=%8.2e,%8.2e)",v0(0),v0(1)));
	if( answer2!="" )
	  sScanF(answer2,"%e %e",&v0(0),&v0(1));
      }
      else
      {
	gi.inputString(answer2,sPrintF(buff,"Enter the initial velocity (default=%8.2e,%8.2e,%8.2e)",
             v0(0),v0(1),v0(2)));
	if( answer2!="" )
	  sScanF(answer2,"%e %e %e",&v0(0),&v0(1),&v0(2));
      }
      v(R,0)=v0(R);
    }
    else if( answer=="initial angular velocity" )
    {
      if( numberOfDimensions==2 )
      {
	gi.inputString(answer2,sPrintF(buff,"Enter the initial angular momentum (default=%8.2e)",w0(2)));
	if( answer2!="" )
	  sScanF(answer2,"%e",&w0(2));
      }
      else
      {
	gi.inputString(answer2,sPrintF(buff,"Enter the initial angular momentum (default=%8.2e,%8.2e,%8.2e)",
             w0(0),w0(1),w0(2)));
	if( answer2!="" )
	  sScanF(answer2,"%e %e %e",&w0(0),&w0(1),&w0(2));
      }
      w(R,0)=w0(R);
    }
    else if( answer=="axes of inertia" )
    {
      if( numberOfDimensions==2 )
      {
	printF("No need to specify axes of inertia in 2D\n");
      }
      else
      {
        for( int axis=0; axis<numberOfDimensions; axis++ )
	{
	  gi.inputString(answer2,sPrintF(buff,"Enter axis %i (default=%8.2e,%8.2e,%8.2e)",
					 axis,e(0,axis),e(1,axis),e(2,axis)));
	  if( answer2!="" )
	    sScanF(answer2,"%e %e %e",&e(0,axis),&e(1,axis),&e(2,axis));
 
          // check that axes are orthogonal and unit length.
          real norm =max(REAL_MIN, SQRT( SQR(e(0,axis))+SQR(e(1,axis))+SQR(e(2,axis)) ));
          e(R,axis)/=norm;
	  for( int dir=0; dir<numberOfDimensions-1; dir++ )
	  {
            real dot = e(0,axis)*e(0,dir)+e(1,axis)*e(1,dir)+e(2,axis)*e(2,dir);
	    if( fabs(dot)>REAL_EPSILON*10. )
	    {
	      printF("ERROR: axis of inertia %i is not orthogonal to axis %i. dot-product=%e \n",axis,dir,dot);
	    }
	  }
	}
      }
    }
    else if( answer=="body force" )
    {
      printF("Specify a const body force to be applied to the rigid body at the center of mass.\n");
	gi.inputString(answer2,sPrintF(buff,"Enter const body force (default=%8.2e,%8.2e,%8.2e)",
             bodyForce(0),bodyForce(1),bodyForce(2)));
	if( answer2!="" )
	  sScanF(answer2,"%e %e %e",&bodyForce(0),&bodyForce(1),&bodyForce(2));

    }
    else if( answer=="position has no constraint" )
    {
      positionConstraint=positionHasNoConstraint;
    }
    else if( answer=="position is constrained to a plane" )
    {
       positionConstraint=positionConstrainedToAPlane;
       for( int k=0; k<=1; k++ )
       {
	 RealArray v(3); 
	 v=0;  v(k)=1.;
	 printF("Enter tangent %i to the constraining plane.\n",k);
	 gi.inputString(answer2,sPrintF(buff,"Enter tangent %i to the constraining plane (default=%8.2e,%8.2e,%8.2e)",
					k,v(0),v(1),v(2)));
	 if( answer2!="" )
	   sScanF(answer2,"%e %e %e",&v(0),&v(1),&v(2));

	 real norm = sqrt(sum(SQR(v)));
	 v/=max(REAL_MIN*100.,norm);
	 constraintValues(Range(3+3*k,5+3*k))=v;
      
	 printF(" Constraining plane: normalized tangent %i is (%7.4f,%7.4,%7.4f)\n",k,v(0),v(1),v(2));
       }
       
    }
    else if( answer=="position is constrained to a line" )
    {
       positionConstraint=positionConstrainedToALine;

      RealArray v(3); 
      v=0;  v(0)=1.;
      printF("Enter the tangent to the constraining line.\n");
      gi.inputString(answer2,sPrintF(buff,"Enter the tangent to the constraining line (default=%8.2e,%8.2e,%8.2e)",
				     v(0),v(1),v(2)));
      if( answer2!="" )
	sScanF(answer2,"%e %e %e",&v(0),&v(1),&v(2));

      real norm = sqrt(sum(SQR(v)));
      v/=max(REAL_MIN*100.,norm);
      constraintValues(Range(3,5))=v;
      
      printF(" Constraining line has a normalized tangent of (%7.4f,%7.4f,%7.4f)\n",v(0),v(1),v(2));
    }
    else if( answer=="position is fixed" )
    {
       positionConstraint=positionIsFixed;
    }
    else if( answer=="rotation has no constraint" )
    {
       rotationConstraint=rotationHasNoConstraint;
    }
    else if( answer=="rotation is constrained to a plane" )
    {
       rotationConstraint=rotationConstrainedToAPlane;
       for( int k=0; k<=1; k++ )
       {
	 RealArray v(3); 
	 v=0;  v(k)=1.;
	 printF("Enter tangent %i to the constraining plane.\n",k);
	 gi.inputString(answer2,sPrintF(buff,"Enter tangent %i to the constraining plane (default=%8.2e,%8.2e,%8.2e)",
					k,v(0),v(1),v(2)));
	 if( answer2!="" )
	   sScanF(answer2,"%e %e %e",&v(0),&v(1),&v(2));

	 real norm = sqrt(sum(SQR(v)));
	 v/=max(REAL_MIN*100.,norm);
	 constraintValues(Range(9+3*k,11+3*k))=v;
      
	 printF(" Constraining plane: normalized tangent %i is (%7.4f,%7.4,%7.4f)\n",k,v(0),v(1),v(2));
       }
    }
    else if( answer=="rotation is constrained to a line" )
    {
       rotationConstraint=rotationConstrainedToALine;
      RealArray v(3); 
      v=0;  v(0)=1.;
      printF("Enter the tangent to the constraining line.\n");
      gi.inputString(answer2,sPrintF(buff,"Enter the tangent to the constraining line (default=%8.2e,%8.2e,%8.2e)",
				     v(0),v(1),v(2)));
      if( answer2!="" )
	sScanF(answer2,"%e %e %e",&v(0),&v(1),&v(2));

      real norm = sqrt(sum(SQR(v)));
      v/=max(REAL_MIN*100.,norm);
      constraintValues(Range(9,11))=v;
      
      printF(" Constraining line has a normalized tangent of (%7.4f,%7.4,%7.4f)\n",v(0),v(1),v(2));
    }
    else if( answer=="rotation is fixed" )
    {
       rotationConstraint=rotationIsFixed;
    }


//      else if( answer=="line constraint" )
//      {
//        constraint=pinnedToALine;

//        printF("Specify the tangent of the line through the center of mass along which the body is "
//               "constrained to move\n");
//        if( constraintValues.getLength(0)==0 )
//        {
//  	constraintValues.redim(3);
//          constraintValues=0.;
//  	constraintValues(0)=1.;
//        }
//        if( numberOfDimensions==2 ) constraintValues(2)=0.;
      
//        gi.inputString(answer2,sPrintF(buff,"Enter the tangent of the constraint line (default=%8.2e,%8.2e,%8.2e)",
//  				     constraintValues(0),constraintValues(1),constraintValues(2)));
//        if( answer2!="" )
//  	sScanF(answer2,"%e %e %e",&constraintValues(0),&constraintValues(1),&constraintValues(2));

//        // normalize the tangent
//        Range Rx=numberOfDimensions;
//        real norm = max( REAL_MIN*100., sqrt( sum(SQR(constraintValues(Rx))) ) );
//        constraintValues/=norm;

//        printF("The body will be constrained to move along the line with tangent (%8.2e,%8.2e,%8.2e)\n",
//             constraintValues(0),constraintValues(1),constraintValues(2));

//      }

    // -- new: commands from a dialog:
    else if( dialog.getToggleValue(answer,"relax correction steps",relaxCorrectionSteps) )
    {
      printF(" The rigid body correction steps can be relaxed. This is often necessary for 'light bodies'.\n"
             " There is a relaxation parameter for the force (alpha) and one for the torque (beta).\n"
             " Both alpha and beta should be a value in [0,1]. As a guess, choose\n"
             " alpha=.5*rho_f/M where rho_f is the density of the surrounding fluid and M is the mass\n"
             " of the rigid body. If the corrections don't converge, choose a smaller value of alpha.\n"
             " The value of beta is related to rho_f/M_I where M_I is a moment of inertial.\n"
             " The iteration for the force is said to converge when \n"
             "   (fDiff/fNorm)<fRelativeTol or fDiff < fAbsoluteTol , where fDiff=|f(new)-f(old)|\n"
             " A similar expression holds for the torque.\n"
	);
    }
    else if( dialog.getTextValue(answer,"debug:","%i",debug) ){}//
    else if( dialog.getTextValue(answer,"force relaxation parameter:","%g",correctionRelaxationParameterForce) ){}//
    else if( dialog.getTextValue(answer,"force relative tol:","%g",correctionRelativeToleranceForce) ){}//
    else if( dialog.getTextValue(answer,"force absolute tol:","%g",correctionAbsoluteToleranceForce) ){}//
    else if( dialog.getTextValue(answer,"torque relaxation parameter:","%g",correctionRelaxationParameterTorque) ){}//
    else if( dialog.getTextValue(answer,"torque relative tol:","%g",correctionRelativeToleranceTorque) ){}//
    else if( dialog.getTextValue(answer,"torque absolute tol:","%g",correctionAbsoluteToleranceTorque) ){}//
    else
    {
      printF("RigidBodyMotion::update:ERROR: unknown response=[%s]\n",(const char*)answer2);
      gi.stopReadingCommandFile();
    }
  }

  gi.unAppendTheDefaultPrompt();
  gi.popGUI(); // restore the previous GUI
  

  return 0;


}

