// ====================================================================================
///  \file trb.C
///  \brief test program to the Newton-Euler rigid body evolution equations (RigidBodyMotion.C)
// ===================================================================================


#include "Overture.h"
#include "SquareMapping.h"
#include "PlotStuff.h"
#include "MatrixTransform.h"
#include "CrossSectionMapping.h"
#include "RigidBodyMotion.h"

  enum TestProblemEnum
  {
    general,
    constantX0Acceleration,
    constantX1Acceleration,
    constantX2Acceleration,
    constantW0Acceleration,
    constantW1Acceleration,
    constantW2Acceleration,
    sinX0Acceleration,
    sinX1Acceleration,
    sinX2Acceleration,
    exponentialV,
    smartExactSolution,
    fallingSphere,
    numberOfTestProblems
  } testProblem=general;

static real lambda=-5.;

static real cf1=2., cf2=1.5, cf3=-1.;
static real cfq1=1., cfq2=2., cfq3=.5;
	  

void
getForce( real t, int component, RealArray & f , RealArray & g, RealArray & xCM, RealArray & vCM )
{
  if( testProblem==constantX0Acceleration|| 
      testProblem==constantX1Acceleration || 
      testProblem==constantX2Acceleration )
  {
    f=0.; 
    f(component)=1.;
    g=0.;
  }
  else if( testProblem==constantW0Acceleration ||
	   testProblem==constantW1Acceleration ||
	   testProblem==constantW2Acceleration )
  {
    f=0.;
    g=0.; 
    g(component)=1.;
      
  }
  else if( testProblem==sinX0Acceleration || 
	   testProblem==sinX1Acceleration ||
	   testProblem==sinX2Acceleration )
  {
    f=0.;
    f(component)=-sin(t);
    g=0.;
  }
  else if( testProblem==general )
  {
    // 
    // 
//     f(0)=sin(Pi*t);
//     f(1)=.5*sin(  Pi*t);
//     f(2)=.5*sin(.5*Pi*t);

//     g(0)=cos(Pi*t);
//     g(1)=cos(.5*Pi*t);
//     g(2)=cos(2.*Pi*t);

    f(0)=cf1*sin(cfq1*Pi*t);
    f(1)=cf2*sin(cfq2*Pi*t);
    f(2)=cf3*sin(cfq3*Pi*t);

    g(0)=     cos(   Pi*t);
    g(1)=1.25*cos(.5*Pi*t);
    g(2)=-.75*cos(2.*Pi*t);
  }
  else if( testProblem==exponentialV )
  {
    f(0)= lambda*vCM(0); 
    f(1)= lambda*vCM(1);
    f(2)= lambda*vCM(2);

    g(0)=0.;
    g(1)=0.;
    g(2)=0.;


  }
  else if( testProblem==smartExactSolution )
  {
    f(0)=0.;
    f(1)=0.;
    f(2)=0.;

    g(0)=0.;
    g(1)=0.;
    g(2)=0.;

  }
  else if( testProblem==fallingSphere )
  {
    // Sphere falling with drag
    // Force = m*g - Cd*.5*rho*v^2   
    //    m=1, Cd=.5
    f(0)=0.;
    f(1)=-1. + .25*SQR(vCM(1));
    f(2)=0.;

    g(0)=0.;
    g(1)=0.;
    g(2)=0.;

  }
  
}



int 
main(int argc, char *argv[]) 
{
  Mapping::debug=0;

  Overture::start(argc,argv);  // initialize Overture and A++/P++

  char buff[180];
  int plotOption=true;
  bool saveMatlabFile=true;  // save results to a matlab file.
  
  real cfl=1., tFinal=5., tPlot=.1;

  int len=0;
  if( argc > 1 )
  { // look at arguments for "-noplot" or "-cfl=<value>"
    aString line;
    for( int i=1; i<argc; i++ )
    {
      line=argv[i];
      if( line=="-noplot" )
        plotOption=false;
      else if( len=line.matches("-cfl=") )
      {
        sScanF(line(len,line.length()-1),"%e",&cfl);
	printf("cfl = %6.2f\n",cfl);
      }
      else if( len=line.matches("-tFinal=") )
      {
        sScanF(line(len,line.length()-1),"%e",&tFinal);
	printf("tFinal = %6.2f\n",tFinal);
      }
    }
  }

  PlotStuff ps(plotOption,"moving");
  PlotStuffParameters psp;
    
  psp.set(GI_PLOT_THE_OBJECT_AND_EXIT,TRUE);
  psp.set(GI_TOP_LABEL,"Original mapping");  // set title
  
  CrossSectionMapping map;
  if( plotOption )
    PlotIt::plot(ps,map,psp);

  RealArray xBound(2,3);
  real xMin=REAL_MAX, xMax=-REAL_MAX;
  for( int axis=0; axis<map.getRangeDimension(); axis++ )
  {
    for( int side=Start; side<=End; side++ )
    {
      Bound b  = map.getRangeBound(side,axis);
      if( b.isFinite() )
      {
	xMin=min(xMin,(real)b);
	xMax=max(xMax,(real)b);
      }
    }
  }
  printf("xMin=%e xMax=%e \n",xMin,xMax);
  
  xBound(Start,Range(0,2))=xMin;
  xBound(End  ,Range(0,2))=xMax;
  
  psp.set(GI_USE_PLOT_BOUNDS,TRUE);  // use the region defined by the plot bounds
  psp.set(GI_PLOT_BOUNDS,xBound); // set plot bounds


  // Use this MatrixTransform to change the existing Mapping, the MatrixTransform
  // can rotate/scale and shift any Mapping
  MatrixTransform transform(map);
  
  // for the Smart exact solution
  real t=0., dt=cfl*.01, t0=0.;
  // const int m0=0, m1=1, m2=2;  // defines a permutation of the exact solution (
  const int m0=1, m1=2, m2=0;  // defines a permutation of the exact solution (
   
  
  int debug=0;

  real mass=1.;           // total mass
  RealArray mI(3);        // 3 moment of inertia
  mI(0)=1.; mI(1)=.5; mI(2)=2.;



  RealArray e(3,3), e0(3,3), r(3,3);      // 3 axes of inertia
  RealArray xCM(3), vCM(3), aCM(3);  // centre of mass: position, velocity
  
  RealArray xNew(3);
  

  RealArray f(3), g(3);  // force and torque

  RealArray omega(3); omega=.0;
  RealArray w(3); w=.0;   // angular velocities about axes of inertia
  RealArray dx(3), vDot(3), dw(3), wDot(3),dTheta(3), eDot(3,3);
  real a,b,c,alpha;

  int numberOfDimensions=map.getRangeDimension();
  RigidBodyMotion body(numberOfDimensions);
  
  body.setProperties(mass,mI,numberOfDimensions);
  
  
//  testProblem=none;
//  testProblem= sinXAcceleration;
//  testProblem=constantW0Acceleration;

  testProblem=constantX0Acceleration;
  
  aString menu[]=
  {
    "general motion",
    "constant x acceleration",
    "constant y acceleration",
    "constant z acceleration",
    "constant omega_x acceleration",
    "constant omega_y acceleration",
    "constant omega_z acceleration",
    "x sine acceleration",
    "y sine acceleration",
    "z sine acceleration",
    "exponential v",
    "smart exact solution",
    "falling sphere",
    "cfl",
    "exit",
    ""
  };
  aString answer;

  for(;;)
  {
    
    testProblem = (TestProblemEnum)ps.getMenuItem(menu,answer,"Choose a test");
    if( answer=="exit" )
      break;
  
    t=0.;
    e(0,0)=1.; e(1,0)=0.; e(2,0)=0.;
    e(0,1)=0.; e(1,1)=1.; e(2,1)=0.;
    e(0,2)=0.; e(1,2)=0.; e(2,2)=1.;
    e0=e;
    w=0.;
    xCM=0.;
    vCM=0.;

    int component=0;
    int mPrint=20;
    if( testProblem==sinX0Acceleration ||
	testProblem==sinX1Acceleration ||
	testProblem==sinX2Acceleration )
    {
      component=(int)testProblem-(int)sinX0Acceleration;
      dt=cfl*.01;
      mPrint= int (20./cfl+.5);
      vCM(component)=+1.;
    }
    else if( testProblem==constantX0Acceleration || 
	     testProblem==constantX1Acceleration || 
	     testProblem==constantX2Acceleration )
    {
      component=(int)testProblem-(int)constantX0Acceleration;
      dt=cfl*.01;
    }
    else if( testProblem==constantW0Acceleration ||
	     testProblem==constantW1Acceleration ||
	     testProblem==constantW2Acceleration )
    {
      component=(int)testProblem-(int)constantW0Acceleration;
    }
    else if( testProblem==exponentialV )
    {
      // solve :  m v' = lambda *v 
      //          x' = v 
      
      lambda=-5.;

      dt=.1*cfl/fabs(lambda);
      
      mPrint= int (20./cfl+.5);

      vCM(0)=1.;
      vCM(1)=2.;
      vCM(2)=3.;
      
      xCM(0)=3.;
      xCM(1)=2.;
      xCM(2)=1.;
      

    }
    else if( testProblem==general )
    {
      // numberOfSteps=2000;
      dt=cfl*.005;
      
//       vCM(0)=-1./(Pi*mass);
//       vCM(1)=-.5/(Pi*mass);
//       vCM(2)=-1./(Pi*mass);
    }
    else if( answer=="cfl" )
    {
      ps.inputString(answer,sPrintF("Enter the cfl number (current=%6.4f)",cfl));
      sScanF(answer,"%e",&cfl);
      printf(" cfl=%6.2f \n",cfl);
      continue;
    }
    else if( testProblem==smartExactSolution )
    {
      // Exact solution from ref E.H. Smart, Advanced Dynamics Vol II
      //    
      //    w1 = a cos(lambda*t)
      //    w2 = b sin(lambda*t)
      //    w3 = c
      
      
      lambda=.25;
      alpha=.5;
      a=1.;  b=-1.;  c=.5;

      dt=.05*cfl; 
      tFinal=5.;
      
      mPrint= 10; 

      vCM(0)=0.;
      vCM(1)=0.;
      vCM(2)=0.;
      
      xCM(0)=0.;
      xCM(1)=0.;
      xCM(2)=0.;
      
      t=0;
      w(m0)=a*cos(lambda*t);
      w(m1)=b*sin(lambda*t);
      w(m2)=c; 
      
      mI(m0)=2.; mI(m1)=2.; 
      mI(m2)=mI(m0)-alpha*mI(m0); 

      body.setProperties(mass,mI,numberOfDimensions);
    }
    else if( testProblem==fallingSphere )
    {
      // Sphere falling to a terminal velocity with a drag force
      
      dt=.05*cfl; 
      tFinal=5.;
      
      mPrint= 9; 

      xCM=0.;
      vCM=0.;
      
    }
    else
    {
      printf("Unknown answer=[%s]\n",(const char*)answer);
      continue;
    }
    
    RealArray xCM0(3), vCM0(3);
    xCM0=xCM;
    vCM0=vCM;
    
    body.setInitialConditions(t,xCM,vCM,w,e0);
    transform.reset();
    
    int numberOfSteps=int(tFinal/dt+1.5); 

    real nextTimeToPlot=tPlot;

    const int numberOfComponents=18;
    RealArray solution;
    if( saveMatlabFile )
    {
      const int maxNumberToSave = int(tFinal/tPlot+2.5);
      solution.redim(maxNumberToSave,numberOfComponents+1); // save t here also
    }
    

    int numberSaved=0;  // for matlab file
    
    for (int i=0; i<=numberOfSteps; i++) 
    {
      getForce( t,component,f,g, xCM,vCM );

      body.integrate( t,f,g, t+dt,xNew,r  );
      body.getVelocity( t+dt,vCM );

      getForce( t+dt,component,f,g, xNew,vCM );
      body.correct( t+dt,f,g, xNew,r  );

      body.getAngularVelocities( t+dt,w );
      body.getVelocity( t+dt,vCM );
      
      // body.getPosition( t+dt,xNew,r );
      
      transform.shift(-xCM(0),-xCM(1),-xCM(2));
      transform.rotate( r );
      transform.shift(xNew(0),xNew(1),xNew(2));
      xCM=xNew;
      t+=dt;

      // body.getAcceleration(t,aCM);
      
      if( (i % mPrint) == 0 )
      {
	if( testProblem==constantX0Acceleration || 
	    testProblem==constantX1Acceleration || 
	    testProblem==constantX2Acceleration )
	  printf("t=%8.2e x=(%7.3e,%7.3e,%7.3e) error(x)=%7.3e error(v)=%7.3e\n",t,xCM(0),xCM(1),xCM(2),
		 fabs(xCM(component)-.5*f(component)/mass*SQR(t)),
                 fabs(vCM(component)- (f(component)/mass*t+vCM0(component))) );
	else if( testProblem==constantW0Acceleration ||
		 testProblem==constantW1Acceleration ||
		 testProblem==constantW2Acceleration )
	  printf("t=%8.2e w=(%7.3e,%7.3e,%7.3e) error=%7.3e \n",t,w(0),w(1),w(2), 
                            fabs(w(component)-g(component)/mI(component)*t));
	else if(  testProblem==sinX0Acceleration || 
		  testProblem==sinX1Acceleration ||
		  testProblem==sinX2Acceleration )
	{
          RealArray x0(3), v0(3);

          x0=0.;
          // here is the true solution
          x0(component) = sin(t)/mass + (vCM0(component) - (cos(t0)/mass))*t + xCM0(component);
	  
          v0=0.;
	  v0(component) = cos(t)/mass+ (vCM0(component) - (cos(t0)/mass));
	  
 	  printf("t=%8.2e xCM=(%8.2e,%8.2e,%8.2e) vCM=(%5.2f,%5.2f,%5.2f) error(x)=(%6.1e,%6.1e,%6.1e) error(v)=(%6.1e,%6.1e,%6.1e) \n",t,
 		 xCM(0),xCM(1),xCM(2),vCM(0),vCM(1),vCM(2),
 		 fabs(xCM(0)-x0(0)),fabs(xCM(1)-x0(1)),fabs(xCM(2)-x0(2)), 
 		 fabs(vCM(0)-v0(0)),fabs(vCM(1)-v0(1)),fabs(vCM(2)-v0(2)));
	}
        else if( testProblem==exponentialV )
	{
          //  v' = lambda*v/m  -->   v = v(0)*exp(lambda/m*t )
          //  x' = v                 x = v(0)*( (m/lambda)*(exp(lambda/m*t)-1.) ) + x(0)
          real err[3], errMax=0.;
          for( int n=0; n<3; n++ )
	  {
            err[n] = fabs( vCM0(n)*( (mass/lambda)*(exp(lambda/mass*t)-1.) ) + xCM0(n)  - xCM(n) );
	    errMax=max(err[n],errMax );
	  }
	  printf("expV: t=%8.2e x=(%7.3e,%7.3e,%7.3e) error=(%7.3e,%7.3e,%7.3e) \n",t,xCM(0),xCM(1),xCM(2),
                 err[0],err[1],err[2]);


	}
        else if( testProblem==general )
	{
// 	  printf("t=%8.2e x=(%7.2e,%7.2e,%7.2e) error(x)=(%6.1e,%6.1e,%6.1e) error(v)=(%6.1e,%6.1e,%6.1e) \n",t,
// 		 xCM(0),xCM(1),xCM(2), 
// 		 fabs(xCM(0)-( -sin(Pi*t)/(Pi*Pi*mass) )),
// 		 fabs(xCM(1)-( -.5*sin(Pi*t)/(Pi*Pi*mass) )),
// 		 fabs(xCM(2)-( -.5*sin(.5*Pi*t)/(.5*Pi*.5*Pi*mass) )),
// 		 fabs(vCM(0)-(    (1.-cos(   Pi*t))/(Pi*mass)   +vCM0(0))),
// 		 fabs(vCM(1)-( .5*(1.-cos(   Pi*t))/(Pi*mass)   +vCM0(1))),
// 		 fabs(vCM(2)-( .5*(1.-cos(.5*Pi*t))/(.5*Pi*mass)+vCM0(2)))
// 	    );

          RealArray x0(3), v0(3);
          // f = cf1*sin(cfq1*t)
          // here is the true solution
          x0(0) = -cf1*sin(cfq1*Pi*t)/(SQR(cfq1*Pi)*mass)  
                    + (vCM0(0) + cf1*cos(cfq1*Pi*t0)/(cfq1*Pi*mass) )*t + xCM0(0);
          x0(1) = -cf2*sin(cfq2*Pi*t)/(SQR(cfq2*Pi)*mass)  
                    + (vCM0(1) + cf2*cos(cfq2*Pi*t0)/(cfq2*Pi*mass) )*t + xCM0(1);
          x0(2) = -cf3*sin(cfq3*Pi*t)/(SQR(cfq3*Pi)*mass) 
                    + (vCM0(2) + cf3*cos(cfq3*Pi*t0)/(cfq3*Pi*mass) )*t + xCM0(2);

          v0(0) =  cf1*(1.-cos(cfq1*Pi*t))/(cfq1*Pi*mass)   +vCM0(0);
          v0(1) =  cf2*(1.-cos(cfq2*Pi*t))/(cfq2*Pi*mass)   +vCM0(1); 
          v0(2) =  cf3*(1.-cos(cfq3*Pi*t))/(cfq3*Pi*mass)   +vCM0(2);

	  

 	  printf("t=%8.2e xCM=(%8.2e,%8.2e,%8.2e) vCM=(%5.2f,%5.2f,%5.2f) error(x)=(%6.1e,%6.1e,%6.1e) error(v)=(%6.1e,%6.1e,%6.1e) \n",t,
 		 xCM(0),xCM(1),xCM(2),vCM(0),vCM(1),vCM(2),
 		 fabs(xCM(0)-x0(0)),fabs(xCM(1)-x0(1)),fabs(xCM(2)-x0(2)), 
 		 fabs(vCM(0)-v0(0)),fabs(vCM(1)-v0(1)),fabs(vCM(2)-v0(2)));
	}
        else if( testProblem==smartExactSolution )
	{
	  // Exact solution from ref E.H. Smart, Advanced Dynamics Vol II
	  //    
	  //    w1 = a cos(lambda*t)
	  //    w2 = b sin(lambda*t)
	  //    w3 = c
          real err[3]; 
          err[m0] = fabs(w(m0)-a*cos(lambda*t));
          err[m1] = fabs(w(m1)-b*sin(lambda*t));
          err[m2] = fabs(w(m2)-c); 

	  printf("smart: t=%8.2e w=(%8.3e,%8.3e,%8.3e) error=(%7.3e,%7.3e,%7.3e) \n",t,w(0),w(1),w(2),
                 err[0],err[1],err[2]);
	}
        else if( testProblem==fallingSphere )
	{
          // Sphere falling with drag

          real err[3]; 
          real x0=0., y0=0., z0=0.;
          real u0=0., v0=0., w0=0.;

          y0= -4.*log(cosh(.5*t)); 
	  v0= -2.*tanh(.5*t); 

	  printf("fallingSphere: t=%8.2e xCM=(%8.3e,%8.3e,%8.3e) error=(%7.3e,%7.3e,%7.3e)\n"
                 "                          vCM=(%8.3e,%8.3e,%8.3e) error=(%7.3e,%7.3e,%7.3e)\n",
                 t,xCM(0),xCM(1),xCM(2),
                 fabs(xCM(0)-x0),fabs(xCM(1)-y0),fabs(xCM(2)-z0),
                 vCM(0),vCM(1),vCM(2),
                 fabs(vCM(0)-u0),fabs(vCM(1)-v0),fabs(vCM(2)-w0));
	}
	
        real dtMax = body.getTimeStepEstimate();
	printf(" ...estimated maximum dt = %9.2e (dt=%9.2e, mass=%6.3f)\n",dtMax,dt,mass);

	// Draw the mapping

        if( plotOption )
	{
	  psp.set(GI_TOP_LABEL,sPrintF(buff,"Mapping at step %i, x=(%4.2f,%4.2f,%4.2f)",i+1,
				       xCM(0),xCM(1),xCM(2)));
//             omega(0)*180./Pi,omega(1)*180./Pi,omega(1)*180./Pi));  // set title

	  psp.set(GI_PLOT_THE_OBJECT_AND_EXIT,TRUE);
	  ps.erase();
	  PlotIt::plot(ps,transform,psp);
	  ps.redraw(TRUE);   // force a redraw
	}
	
      }
      

      if( saveMatlabFile && t>nextTimeToPlot-.5*dt )
      {
 	nextTimeToPlot+=tPlot;
 	int i=numberSaved;
	  
 	assert( i<solution.getBound(0) );
	  
 	solution(i,0)=t;

 	solution(i,1)=xCM(0);
 	solution(i,2)=xCM(1);
 	solution(i,3)=xCM(2);

 	solution(i,4)=vCM(0);
 	solution(i,5)=vCM(1);
 	solution(i,6)=vCM(2);

 	solution(i,7)=w(0);
 	solution(i,8)=w(1);
 	solution(i,9)=w(2);
	  
 	RealArray e(3,3);
 	body.getAxesOfInertia( t,e );
	  
 	solution(i,10)=e(0,0);
 	solution(i,11)=e(1,0);
 	solution(i,12)=e(2,0);
			    			    
 	solution(i,13)=e(0,1);
 	solution(i,14)=e(1,1);
 	solution(i,15)=e(2,1);
			    			    
 	solution(i,16)=e(0,2);
 	solution(i,17)=e(1,2);
 	solution(i,18)=e(2,2);

//  	solution(i,10)=e(0,0);
//  	solution(i,11)=e(0,1);
//  	solution(i,12)=e(0,2);

//  	solution(i,13)=e(1,0);
//  	solution(i,14)=e(1,1);
//  	solution(i,15)=e(1,2);

//  	solution(i,16)=e(2,0);
//  	solution(i,17)=e(2,1);
//  	solution(i,18)=e(2,2);
	  
 	numberSaved++;
	  
      }

    

    } // end for (i )  (time steps)

    if( saveMatlabFile )
    {
      
      FILE *matlabFile=NULL;
      aString fileName="rigidBody.m";
      // gi.inputString(fileName,sPrintF(answer,
      //     "Enter the name of the matlab file (default=%s)\n",(const char*)fileName));

      matlabFile = fopen((const char*)fileName,"w" ); 
    
      const int numPerLine=5;
      fprintf(matlabFile,"tc=[");
      for( int i=0; i<numberSaved; i++ ) 
      {
	fprintf(matlabFile,"%17.10e ",solution(i,0));
	if( (i % numPerLine)==numPerLine-1 ) fprintf(matlabFile,"...\n");
      }
      fprintf(matlabFile,"];\n");


      fprintf(matlabFile,"yc=[");
      for( int i=0; i<numberSaved; i++ ) 
      {
	for( int j=1; j<=numberOfComponents; j++ )
	{
	  fprintf(matlabFile,"%17.10e ",solution(i,j));
	  // if( (i % numPerLine)==numPerLine-1 ) fprintf(matlabFile,"...\n");
	}
        fprintf(matlabFile," ; \n");  // end this row  solution(1:nc,i)
      }
      
      fprintf(matlabFile,"];\n");
      
      fclose(matlabFile);
  
      printf("Saved file %s for matlab\n",(const char*)fileName);
      
    }
    


  } // for(;;)
  
  Overture::finish(); 
  return 0;
}
