#include "ProbeInfo.h"
#include "GenericGraphicsInterface.h"
#include "HDF_DataBase.h"
// #include "GUIState.h"

// we need:
//    frequencyToSaveProbes

// ==========================================================================================
/// \brief : define a probe. A probe is a point or location in a grid where the
///    solution is periodically saved to a file. A probe may also save the solution on a 
///   bounding-box. This boudning box info may be used on subsequent computations as 
///   boundary conditions for a computation on a smaller sub-domain.
/// 
///  \details The file generated with the probe results can be read by the matlab script
///  plotProbes.m 
// ==========================================================================================
ProbeInfo::
ProbeInfo()
{
  probeType=probeAtGridPoint;
  file=NULL;
  fileName="probeFile.dat";  
  pdb=NULL;
  
  grid=0;
  iv[0]=iv[1]=iv[2]=0;

  xv[0]=xv[1]=xv[2]=0.;

  #define bb(side,axis) boundingBox[(side)+2*(axis)]
  boundingBoxGrid=0;
  for( int axis=0; axis<3; axis++ ) 
  {
    bb(0,axis)=0;
    bb(1,axis)=0;
  }
  
  numberOfLayers=2;
  numberOfTimes=0;
  times=NULL;
}

ProbeInfo::
~ProbeInfo()
{
  if( file!=NULL )
  {
    fclose(file);
  }
  // close the data base file for the bounding box info
  if( pdb!=NULL )
  {
    printF("++++ ProbeInfo: close the bounding box probe file +++\n");
    // save the time history:
    pdb->put(numberOfTimes,"numberOfTimes");
    if( numberOfTimes>0 )
    {
      assert( times!=NULL );
      pdb->put((*times)(Range(numberOfTimes)),"times");
    }
    delete times;
    pdb->unmount();
    delete pdb;

    HDF_DataBase db;
    db.mount("cylProbeBox.hdf","R");
    int nt=-1;
    db.get(nt,"numberOfTimes");
    printF("Re-mount the probe file, numberOfTimes=%i\n",nt);
    

  }
  
}

// ===========================================================================================
/// \brief Compute the closest grid point to each of a set of points in space.
/// \details : this function is used for examples to compute the closest grid points
///    for probes. 
/// \param points(0:numPoints,0:d-1) (input) : coordinates of points to check.
/// \param gridLocation(0:numPoints,0:3) (output) : (i1,i2,i3,grid) closest grid and point. 
/// 
// ===========================================================================================
//#include "InterpolatePoints.h"
#include "UnstructuredMapping.h"
int
getClosestGridPoint( CompositeGrid & cg, 
                     RealArray & points, 
                     IntegerArray & gridLocation )
{
  
  int numberOfPoints=points.getLength(0);

  gridLocation.redim(numberOfPoints,4);

  IntegerArray indexValues, interpoleeGrid;

  if( false )
  {
    // ***fix**** this requires the center array I think
    //InterpolatePoints interp;
    //interp.buildInterpolationInfo(positionToInterpolate,cg );
    //interp.getInterpolationInfo(cg, indexValues, interpoleeGrid);
  }
  else
  {
    // locate the nearest grid point
          
    const int numberOfDimensions = cg.numberOfDimensions();
    Range I=numberOfPoints;
    RealArray r(I,numberOfDimensions);
    RealArray x(I,numberOfDimensions);
    for( int i=0; i<numberOfPoints; i++ )
    {
      for( int axis=0; axis<numberOfDimensions; axis++ )
      {
	x(i,axis)=points(i,axis);
      }
    }
	  
    indexValues.redim(I,numberOfDimensions);
    interpoleeGrid.redim(I);
    interpoleeGrid=-1;

    int iv[3], &i1=iv[0], &i2=iv[1], &i3=iv[2];
    int numFound=0;
    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      Mapping & map = mg.mapping().getMapping();
      const intArray & mask = mg.mask();
      i3=mg.gridIndexRange(0,2);

      if ( mg.getGridType()==MappedGrid::structuredGrid )
      {
	r=-1.;
        #ifdef USE_PPP
	  map.inverseMapS(x,r);
        #else
	  map.inverseMap(x,r);
        #endif 
	for( int i=0; i<numberOfPoints; i++ )
	{
	  if( interpoleeGrid(i)<0 ) // this point not yet found
	  {
	    bool ok=true;
	    for( int axis=0; axis<numberOfDimensions; axis++ )
	    {
	      // closest point:
	      iv[axis] = int( r(i,axis)/mg.gridSpacing(axis)+ mg.gridIndexRange(0,axis) +.5);
	      if( iv[axis]< mg.gridIndexRange(0,axis) ||
		  iv[axis]> mg.gridIndexRange(1,axis) )
	      {
		ok=false;
		break;
	      }
	    }
#ifndef USE_PPP
	    if( ok && mask(i1,i2,i3)>0 ) // *** fix this -- P++ problem --
#else
	      if( ok ) 
#endif
	      {
		for( int axis=0; axis<numberOfDimensions; axis++ )
		  indexValues(i,axis)=iv[axis];
		interpoleeGrid(i)=grid;
		numFound++;
	      }
	  }
		
	} // end for i
      }
      else
      {
        // --- Unstructured grid ---
	for( int i=0; i<numberOfPoints; i++ )
	{
	  real xx[3];
	  xx[2] = 0.;
	  for ( int a=0; a<mg.numberOfDimensions(); a++ )
	    xx[a] = x(i,a);

	  UnstructuredMapping & umap = (UnstructuredMapping &)mg.mapping().getMapping();
		    
	  int ent = umap.findClosestEntity(UnstructuredMapping::Face, xx[0],xx[1],xx[2]);
	  assert(ent!=-1);

	  interpoleeGrid(i) = grid;
	  indexValues(i,0) = ent;
	  for ( int a=1; a<mg.numberOfDimensions(); a++ )
	    indexValues(i,a) = 0;
	  numFound++;
	}
      }

      if( numFound==numberOfPoints ) break;
    } // end for grid
  }
	
	
  for( int i=0,j=0; i<numberOfPoints; i++ )
  {
    gridLocation(i,0)=indexValues(i,0);
    gridLocation(i,1)=indexValues(i,1);
    if( cg.numberOfDimensions()==3 )
      gridLocation(i,2)=indexValues(i,2);
    else
      gridLocation(i,2)=0;
	  
    gridLocation(i,3)=interpoleeGrid(i);
    if( interpoleeGrid(i)<0 )
    {
      printF("getClosestGridPoint: error location points %i: x=(%9.3e,%9.3e,%9.3e)\n",i,
	     points(i,0),points(i,1),points(i,2));
      Overture::abort();
    }

    printF("getClosestGridPoint: i=%i: x=(%9.3e,%9.3e,%9.3e), closest grid=%i, pt i=(%i,%i,%i)\n",i,
	   points(i,0),points(i,1),points(i,2),
	   gridLocation(i,3),gridLocation(i,0),gridLocation(i,1),gridLocation(i,2)   );
  }
  
  return 0;
}



// ===========================================================================================
/// \brief define the properties of the probe
///
/// \params cg (input) : the probe will be a point on this grid
// ===========================================================================================
int ProbeInfo::
update( CompositeGrid & cg, GenericGraphicsInterface & gi )
{
  int returnValue=0;
  
  GUIState gui;
  gui.setWindowTitle("Probe Data");
  gui.setExitCommand("exit", "continue");
  DialogData & dialog = (DialogData &)gui;

  aString prefix = ""; // prefix for commands to make them unique.

  if( true )
  {

    const int maxCommands=40;
    aString cmd[maxCommands];

//     aString label[] = {"compressed", "uncompressed","" }; //
//     addPrefix(label,prefix,cmd,maxCommands);
//     dialog.addOptionMenu("mode", cmd, label, ( dbase.get<int >("useStreamMode")? 0 : 1));

//     aString tbCommands[] = {"compressed",""};
//     int tbState[10];
//     tbState[0] = 0; // parameters.dbase.get<bool >("twilightZoneFlow");
//     tbState[1] = 0; // parameters.dbase.get<bool >("projectInitialConditions");
//     int numColumns=1;
//     dialog.setToggleButtons(tbCommands, tbCommands, tbState, numColumns); 


    // aString pbLabels[] = {"save a bounding box",""};
    // addPrefix(pbLabels,prefix,cmd,maxCommands);
    // int numRows=1;
    // dialog.setPushButtons( cmd, pbLabels, numRows ); 

    const int numberOfTextStrings=40;
    aString textLabels[numberOfTextStrings];
    aString textStrings[numberOfTextStrings];


    int nt=0;
    
    textLabels[nt] = "grid point";  sPrintF(textStrings[nt], "%i %i %i %i (grid,i1,i2,i3)",grid,iv[0],iv[1],iv[2]); nt++;
    textLabels[nt] = "nearest grid point to"; sPrintF(textStrings[nt], "%8.2e %8.2e %8.2e",0.,0.,0.); nt++;
														      

    textLabels[nt] = "bounding box grid";  sPrintF(textStrings[nt], "%i",boundingBoxGrid); nt++; 
    
    textLabels[nt] = "bounding box";  sPrintF(textStrings[nt], "%i %i, %i %i, %i %i (i1a,i1b, i2a,i2b, i3a,i3b)",
					      bb(0,0),bb(1,0), bb(0,1), bb(1,1), bb(0,2), bb(1,2)); nt++;
    
    textLabels[nt] = "number of layers";  sPrintF(textStrings[nt], "%i",numberOfLayers);  nt++; 

    textLabels[nt] = "file name"; sPrintF(textStrings[nt], "%s",(const char*)fileName);  nt++; 


    // null strings terminal list
    textLabels[nt]="";   textStrings[nt]="";  assert( nt<numberOfTextStrings );
    // addPrefix(textLabels,prefix,cmd,maxCommands);
    // dialog.setTextBoxes(cmd, textLabels, textStrings);
    dialog.setTextBoxes(textLabels, textLabels, textStrings);

  }
  

  aString answer,answer2;
  
  gi.pushGUI(gui);
  gi.appendToTheDefaultPrompt("probe>");  

  int len=0;
  for(int it=0; ; it++)
  {
    gi.getAnswer(answer,"");
  
    if( answer(0,prefix.length()-1)==prefix )
      answer=answer(prefix.length(),answer.length()-1);

    if( answer=="exit" || answer=="done" )
    {
      break;
    }
    else if( len=answer.matches("grid point") )
    {
      probeType=probeAtGridPoint;
      
      sScanF(answer(len,answer.length()-1),"%i %i %i %i",&grid,&iv[0],&iv[1],&iv[2]);
      printF("ProbeInfo:Using grid point (grid,i1,i2,i3)=(%i,%i,%i)\n",grid,iv[0],iv[1],iv[2]);
      dialog.setTextLabel("grid point",sPrintF(answer2,"%i %i %i %i (grid,i1,i2,i3)",grid,iv[0],iv[1],iv[2]));
    }
    else if( len=answer.matches("nearest grid point to") )
    {
      probeType=probeAtLocation;
      
      RealArray x(1,3);
      IntegerArray gridLocation(1,4);
      sScanF(answer(len,answer.length()-1),"%e %e %e",&x(0,0),&x(0,1),&x(0,2));
      dialog.setTextLabel("nearest grid point to",sPrintF(answer2,"%8.2e %8.2e %8.2e",x(0,0),x(0,1),x(0,2)));

      printF("ProbeInfo: finding closest grid point to x=(%e,%e,%e)\n",x(0,0),x(0,1),x(0,2));
      getClosestGridPoint( cg,x,gridLocation );
      if( gridLocation(0,3)>=0 )
      {
	iv[0]=gridLocation(0,0);
	iv[1]=gridLocation(0,1);
	iv[2]=gridLocation(0,2);
	grid=gridLocation(0,3);
	printF("ProbeInfo:Using grid=%i, (i1,i2,i3)=(%i,%i,%i)\n",grid,iv[0],iv[1],iv[2]);
	if( true )
	{
	  int i1=iv[0], i2=iv[1], i3=iv[2];
	  cg[grid].update(MappedGrid::THEcenter);
	  realArray & center = cg[grid].center();
	  real dist = SQR(x(0,0)-center(i1,i2,i3,0)) + SQR(x(0,1)-center(i1,i2,i3,1));
	  if( cg.numberOfDimensions()==3 ) dist += SQR(x(0,2)-center(i1,i2,i3,2));
	  dist=sqrt(dist);
	  printF("ProbeInfo: pt=(%e,%e,%e), grid-pt=(%e,%e,%e), dist=%8.2e\n",x(0,0),x(0,1),x(0,2),
                 center(i1,i2,i3,0),center(i1,i2,i3,1),(cg.numberOfDimensions()==3 ? center(i1,i2,i3,2) : 0.),
                 dist);
	}
      }
      else
      {
	printF("ProbeInfo:ERROR finding closest grid point to x=(%e,%e,%e)\n",x(0,0),x(0,1),x(0,2));
        
      }
    }
    else if( dialog.getTextValue(answer,"bounding box grid","%i",boundingBoxGrid) )
    {
      probeType=probeBoundingBox;
      if( boundingBoxGrid>=0 && boundingBoxGrid<cg.numberOfComponentGrids() )
      {
	const IntegerArray & gid = cg[boundingBoxGrid].gridIndexRange();
	printF("ProbeInfo: The grid index range on grid=%i (%s) is [%i,%i]x[%i,%i]x[%i,%i]\n",
	       boundingBoxGrid, (const char*)cg[boundingBoxGrid].getName(),
               gid(0,0),gid(1,0),gid(0,1),gid(1,1),gid(0,2),gid(1,2));
      }
      else
      {
	printF("ProbeInfo:ERROR: invalid choice for grid=%i. There are only %i grids\n",
	       boundingBoxGrid,cg.numberOfComponentGrids());
	boundingBoxGrid=0;
	dialog.setTextLabel("ounding box grid",sPrintF("%i",boundingBoxGrid));
      }
    }
    else if( len=answer.matches("bounding box") )
    {
      probeType=probeBoundingBox;

      sScanF(answer(len,answer.length()-1),"%i %i %i %i %i %i",&bb(0,0),&bb(1,0), &bb(0,1), &bb(1,1), 
             &bb(0,2), &bb(1,2));
      if( boundingBoxGrid>=0 && boundingBoxGrid<cg.numberOfComponentGrids() )
      {
	const IntegerArray & gid = cg[boundingBoxGrid].gridIndexRange();
	for( int axis=0; axis<3; axis++ )
	for( int side=0; side<=1; side++ )
	{
	  if( (side==0 && bb(0,axis)<gid(0,axis)) || ( side==1 && bb(1,axis)>gid(1,axis)) )
	  {
	    printF("ProbeInfo: error: bb(%i,%i)=%i is outside the gridIndexRange(%i,%i)=%i. Setting to %i\n",
                   side,axis,bb(side,axis), side,axis,gid(side,axis), gid(side,axis));
	    bb(side,axis)=gid(side,axis);
	  }
	}
	
      }
      
      printF("ProbeInfo: Saving the solution from grid %i (%s) on the bounding box:\n"
             "  [i1a,i1b]x[i2a,i2b]x[i3a,i3b]=[%i,%i]x[%i,%i]x[%i,%i]\n"
             "  You should also set the `number of layers'. Current number of layers=%i\n",
	     boundingBoxGrid, (const char*)cg[boundingBoxGrid].getName(),
             bb(0,0),bb(1,0), bb(0,1), bb(1,1), bb(0,2), bb(1,2), numberOfLayers);
      dialog.setTextLabel("bounding box",sPrintF(answer2,"%i %i, %i %i, %i %i (i1a,i1b, i2a,i2b, i3a,i3b)",
						 bb(0,0),bb(1,0), bb(0,1), bb(1,1), bb(0,2), bb(1,2)));
    }
    else if( dialog.getTextValue(answer,"number of layers","%i",numberOfLayers) ){} // 

    else if( len=answer.matches("file name") )
    {
      // remove initial and trailing blanks.
      int length=answer.length();
      int istart=len;
      while( istart<length && answer[istart]==' ') istart++;
      int iend=length-1;
      while( iend>=istart && answer[iend]==' ') iend--;
      if( iend>=istart )
      {
	fileName=answer(istart,iend);
	printF("ProbeInfo::update: setting probe file name = [%s]\n",(const char*)fileName);
      }
      else
      {
        printF("ProbeInfo::update:ERROR: file name was empty!\n");
        gi.stopReadingCommandFile();
      }
    }
    else
    {
      printF("ProbeInfo::update:ERROR: Unknown response: [%s]\n",(const char*)answer);
      gi.stopReadingCommandFile();
       
    }
    
  }

  if( true ) // !executeCommand 
  {
    gi.popGUI();
    gi.unAppendTheDefaultPrompt();
  }
  
  return returnValue;

}
