#include "DomainSolver.h"
#include "CompositeGridOperators.h"
#include "Oges.h"
#include "LineSolve.h"
#include "Regrid.h"
#include "ParallelUtility.h"
#include <time.h>

// ===================================================================================================================
/// \brief Solve equations to time tFinal,
// ===================================================================================================================
int DomainSolver::
solve()    
//===================================================================================
// /Description:
// solve equations to time tFinal
//
// /Author: WDH
//
//\end{DomainSolverInclude.tex}  
// =======================================================================================================
{
  real cpu0 = getCPU();
  
  advance(parameters.dbase.get<real >("tFinal"));

//   cg.reference(compositeGridSolver[0]->solution.cg);
//   u.reference(compositeGridSolver[0]->solution.u);
  
  parameters.dbase.get<RealArray>("timing")(Parameters::totalTime)+=getCPU()-cpu0;
  return 0;
}

// ===================================================================================================================
/// \brief Advance one time-step. This function is used by the multi-physics solver Cgmp.
/// \param t (input) : current time.
/// \param dt (input) : time step.
/// \param stepNumber (input) : current counter for the step number.
/// \param numberOfSubSteps (input) : number of sub-steps to take.
// ===================================================================================================================
void DomainSolver::
takeOneStep( real & t, real & dt, int stepNumber, int & numberOfSubSteps )
{
  Parameters::TimeSteppingMethod & timeSteppingMethod = 
                       parameters.dbase.get<Parameters::TimeSteppingMethod>("timeSteppingMethod");

  if( !parameters.dbase.get<DataBase >("modelData").has_key("initializeAdvance") )
    parameters.dbase.get<DataBase >("modelData").put<int>("initializeAdvance",true);
  int & init=parameters.dbase.get<DataBase >("modelData").get<int>("initializeAdvance");


  if( timeSteppingMethod==Parameters::forwardEuler )
  {
    const int next = (current+1) %2;

    GridFunction & gf0 = gf[current];
    GridFunction & gf1 = gf[next];
    realCompositeGridFunction & fn0 = fn[0];
	  
    eulerStep(t,t,t+dt,dt,gf0,gf0,gf1,fn0,fn0,stepNumber  ,numberOfSubSteps);
  }
  else if( timeSteppingMethod==Parameters::adamsBashforth2 ||
	   timeSteppingMethod==Parameters::adamsPredictorCorrector2 ||
	   timeSteppingMethod==Parameters::adamsPredictorCorrector4 )
  {
    advanceAdamsPredictorCorrector( t,dt, numberOfSubSteps,init,stepNumber ); 
  }
  else if( timeSteppingMethod==Parameters::implicit )
  {
    if ( parameters.dbase.get<Parameters::ImplicitMethod >("implicitMethod")==Parameters::trapezoidal )
      advanceTrapezoidal(t,dt,numberOfSubSteps,init,stepNumber);
    else
      advanceImplicitMultiStep( t,dt, numberOfSubSteps,init,stepNumber );

  }
  else
  {
    printF("DomainSolver::takeOneStep:ERROR: timeSteppingMethod=%i is not yet implemented\n",(int)timeSteppingMethod);
    Overture::abort("DomainSolver::takeOneStep:ERROR");
  }

}




// ===================================================================================================================
/// \brief Set the time to integrate to.
/// \param tFinal_ (input) : integrate to this time this final time.
// ===================================================================================================================
int DomainSolver::
setFinalTime(const real & tFinal_ )
{
  parameters.dbase.get<real >("tFinal")=tFinal_;
  return 0;
}



// ===================================================================================================================
/// \brief Output timing statistics. 
/// \details This information is normally printed at the end of the run.
/// \param file (input) : output to this file.
// ===================================================================================================================
int DomainSolver::
printStatistics(FILE *file /* = stdout */)
{
  const int np= max(1,Communication_Manager::numberOfProcessors());
  const int myid=max(0,Communication_Manager::My_Process_Number);
  fflush(0);
  Communication_Manager::Sync();
  
  FILE *& logFile =parameters.dbase.get<FILE* >("logFile");

  fPrintF(logFile,"\n"
	  " -------------- Final Statistics for %s  ---------------\n",(const char*)getClassName());

  if( np>1 )
  { // output parallel distribution: 
    aString buff;
    cg.displayDistribution(sPrintF(buff,"%s summary",(const char*)getClassName()),logFile);
  }
  
  if( poisson!=0 )
  {
    fPrintF(logFile,"\n"
	    " ----------------- Poisson Solver -------------------------------------- \n");
    poisson->printStatistics(logFile);
  }
  for( int imp=0; imp<numberOfImplicitSolvers; imp++ )
  {
    fPrintF(logFile,"\n"
	    "&&&&&&&&&&&&&&&&&&&&&&&&&&& Implicit Solver %i &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& \n",imp);
    implicitSolver[imp].printStatistics(logFile);
    fPrintF(logFile,"&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n");
  }

  
  // output timings from Interpolant.
  Interpolant::printStatistics(logFile);

  printMemoryUsage(logFile); // note: communication required here

  GenericMappedGridOperators::printBoundaryConditionStatistics(logFile);

  if( np>1 && parameters.isAdaptiveGridProblem() && parameters.dbase.get<Regrid* >("regrid")!=NULL )
  { // print load balancing statistics
    parameters.dbase.get<Regrid* >("regrid")->getLoadBalancer().printStatistics(logFile);
  }
  

  // const int numSteps=max(1,numberOfStepsTaken);
  const int numSteps=max(1,parameters.dbase.get<int >("globalStepNumber"));

  RealArray & timing = parameters.dbase.get<RealArray >("timing");
  //  const aString *timingName = parameters.dbase.get<ArraySimpleFixed<aString,Parameters::maximumNumberOfTimings,1,1,1> >("timingName");
//  const ArraySimpleFixed<aString,Parameters::maximumNumberOfTimings,1,1,1> &timingName = parameters.dbase.get<ArraySimpleFixed<aString,Parameters::maximumNumberOfTimings,1,1,1> >("timingName");

  const std::vector<aString> & timingName = parameters.dbase.get<std::vector<aString> >("timingName");
  const int & maximumNumberOfTimings = Parameters::maximumNumberOfTimings;

  const Parameters::TimeSteppingMethod & timeSteppingMethod = 
           parameters.dbase.get<Parameters::TimeSteppingMethod >("timeSteppingMethod");
  
  // count the total number of grid points
  real totalNumberOfGridPoints=0.;
  int grid;
  if( parameters.dbase.get<real >("numberOfRegrids")==0. )
  {
    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
      totalNumberOfGridPoints+=numberOfGridPoints[grid];
  }
  else
  {
    totalNumberOfGridPoints=parameters.dbase.get<real >("sumTotalNumberOfGridPoints")/parameters.dbase.get<real >("numberOfRegrids");  // use average
  }
  
  
  // adjust times for waiting
  real timeForWaiting=timing(Parameters::timeForWaiting);
  timing(Parameters::timeForPlotting)-=timeForWaiting;
  timing(Parameters::totalTime)-=timeForWaiting;
  timing(Parameters::timeForAdvance)-=timeForWaiting;

  // Compute timeForOther -- times we have not counted
  real totalMeasured=0;  // sum of times we have measured
  for( int i=0; i<Parameters::maximumNumberOfTimings; i++ )
  {
//     if( i!=Parameters::totalTime && 
//         i!=Parameters::timeForAdvance &&    // timeForAdvance is split into sub-categories so don't count 
//         i!=Parameters::timeForAmrRegrid &&              // timeForRegrid is split into sub-categories so don't count 
//         i!=Parameters::timeForOther && 
//         i!=Parameters::timeForWaiting )
    // Add up the cpu times we have measured -- exclude total time and any names that beging with a blank
    // since these are assumed to be sub-times of a larger category
    if( i!=Parameters::totalTime && timingName[i]!="" && timingName[i][0]!=' ' )
      totalMeasured+=timing(i);
  }
  timing(Parameters::timeForOther)=timing(Parameters::totalTime)-totalMeasured;
  printF("\n ****  CPU: totalTime=%9.3e totalMeasured=%9.3e, timeForOther=%8.2e \n",
          timing(Parameters::totalTime),totalMeasured,timing(Parameters::timeForOther));
  

  timing(Parameters::totalTime)=max(timing(Parameters::totalTime),REAL_MIN*10.);   // total time 

  // Get max/ave times 
  RealArray maxTiming(timing.dimension(0)),minTiming(timing.dimension(0)),aveTiming(timing.dimension(0));

//   for( int i=0; i<Parameters::maximumNumberOfTimings; i++ )
//     maxTiming(i)=ParallelUtility::getMaxValue(timing(i));   // max over all processors  -- is this the right thing to do?
  
  ParallelUtility::getMaxValues(&timing(0),&maxTiming(0),Parameters::maximumNumberOfTimings);
  ParallelUtility::getMinValues(&timing(0),&minTiming(0),Parameters::maximumNumberOfTimings);
  ParallelUtility::getSums(&timing(0),&aveTiming(0),Parameters::maximumNumberOfTimings);
  aveTiming/=np;

  real mem=Overture::getCurrentMemoryUsage();
  real maxMem=ParallelUtility::getMaxValue(mem);  // max over all processors
  real minMem=ParallelUtility::getMinValue(mem);  // min over all processors
  real totalMem=ParallelUtility::getSum(mem);  // min over all processors
  real aveMem=totalMem/np;
  real maxMemRecorded=ParallelUtility::getMaxValue(Overture::getMaximumMemoryUsage());
  

  // **********************
  if( myid==0 )
  {

    // Get the current date
    time_t *tp= new time_t;
    time(tp);
    // tm *ptm=localtime(tp);
    const char *dateString = ctime(tp);

    for( int fileio=0; fileio<2; fileio++ )
    {
      FILE *output = fileio==0 ? logFile : file;

      if( parameters.dbase.get<real >("numberOfRegrids")==0 )
      {
	fprintf(output,
                "\n"
                "         ---%s Summary : %s --- \n"
                "            %s"          
		"  ==== numberOfStepsTaken =%i, number of grids=%i, number of gridpts =%g, processors=%i ==== \n"
		"  ==== memory per-proc: [min=%g,ave=%g,max=%g](Mb), max-recorded=%g (Mb), total=%g (Mb)\n"
		"   Timings:           (ave-sec/proc:) seconds    sec/step   sec/step/pt     %%     [max-s/proc] [min-s/proc]\n",
		(const char*)getClassName(),(const char*)getName(),
                dateString,
		numSteps,cg.numberOfComponentGrids(),totalNumberOfGridPoints,np,
                minMem,aveMem,maxMem,maxMemRecorded,totalMem);
      }
      else
      {
	int levels = 1, ratio=1;
	aString loadBalanceName="off";
	if( parameters.dbase.get<Regrid* >("regrid")!=NULL )
	{
	  levels=parameters.dbase.get<Regrid* >("regrid")->getDefaultNumberOfRefinementLevels();
	  ratio=parameters.dbase.get<Regrid* >("regrid")->getRefinementRatio();
	  if( parameters.dbase.get<Regrid* >("regrid")->loadBalancingIsOn() )
	    loadBalanceName=parameters.dbase.get<Regrid* >("regrid")->getLoadBalancer().getLoadBalancerTypeName();
	}
      
	fprintf(output,
                "\n"
                "         ---%s Summary : %s --- \n"
                "           %s"          
		"  ==== numberOfStepsTaken =%i, processors=%i\n"
		"  ==== AMR levels=%i, AMR ratio=%i, load-balance=%s\n"
		"  ==== number of grids:   [min=%g,max=%g,ave=%g] (number of regrid steps=%g)\n"
		"  ==== number of gridpts: [min=%g,max=%g,ave=%g], \n"
		"  ==== memory per-proc: [min=%g,ave=%g,max=%g](Mb), max-recorded=%g (Mb), total=%g (Mb)\n"
		"   Timings:           (ave-sec/proc:) seconds    sec/step   sec/step/pt     %%     [max-s/proc] [min-s/proc]\n",
		(const char*)getClassName(),(const char*)getName(),
                dateString,
		numSteps,np,
		levels,ratio,(const char*)loadBalanceName,
		parameters.dbase.get<real >("minimumNumberOfGrids"),
		parameters.dbase.get<real >("maximumNumberOfGrids"),
		parameters.dbase.get<real >("totalNumberOfGrids")/parameters.dbase.get<real >("numberOfRegrids"),
		parameters.dbase.get<real >("numberOfRegrids"),
		parameters.dbase.get<real >("minimumNumberOfGridPoints"),parameters.dbase.get<real >("maximumNumberOfGridPoints"),
		parameters.dbase.get<real >("sumTotalNumberOfGridPoints")/parameters.dbase.get<real >("numberOfRegrids"),
		minMem,aveMem,maxMem,maxMemRecorded,totalMem);
      }
    
  
      int nSpace=35;
      aString dots="........................................................................";
      if( maxTiming(0)==0. )
	maxTiming(0)=REAL_MIN;
      for( int i=0; i<Parameters::maximumNumberOfTimings; i++ )
	if( timingName[i]!="" && aveTiming(i)!=0. )    
	  fprintf(output,"%s%s%10.3e  %10.3e  %10.3e   %7.3f  %10.3e  %10.3e\n",(const char*)timingName[i],
		  (const char*)dots(0,max(0,nSpace-timingName[i].length())),
		  aveTiming(i),aveTiming(i)/numSteps,aveTiming(i)/numSteps/totalNumberOfGridPoints,
		  100.*aveTiming(i)/aveTiming(Parameters::totalTime),maxTiming(i),minTiming(i));

      if( timeSteppingMethod==Parameters::implicit )
	fprintf(output,"implicit time stepping: average number of iterations to solve implicit system =%5.1f/step\n",
		real(parameters.dbase.get<int >("numberOfIterationsForImplicitTimeStepping"))/max(1.,numSteps));

      if( timeSteppingMethod==Parameters::implicit || 
          timeSteppingMethod==Parameters::adamsPredictorCorrector2 || 
          timeSteppingMethod==Parameters::adamsPredictorCorrector4)
      fprintf(output,"Predictor-correction time stepping: Average number of corrector steps =%5.2f.\n",
	      real(parameters.dbase.get<int>("totalNumberOfPCcorrections"))/max(1.,numSteps));


      if( poisson )//parameters.dbase.get<Parameters::PDE >("pde")==Parameters::incompressibleNavierStokes ||
	//	  parameters.dbase.get<Parameters::PDE >("pde")==Parameters::allSpeedNavierStokes )
	fprintf(output,"pressure equation: average number of iterations to solve =%f/solve (%3.2f solves/step)\n",
		real(parameters.dbase.get<int >("numberOfIterationsForConstraints"))/max(1,parameters.dbase.get<int >("numberOfSolvesForConstraints")),
		real(parameters.dbase.get<int >("numberOfSolvesForConstraints"))/max(1,numSteps));
    
#ifdef USE_PPP
      fprintf(output," Total: messages sent=%i messages received=%i\n",
              Diagnostic_Manager::getNumberOfMessagesSent(),Diagnostic_Manager::getNumberOfMessagesReceived());
      fprintf(output," Per step: messages sent=%6.1f messages received=%6.1f\n",
              Diagnostic_Manager::getNumberOfMessagesSent()/max(1.,numSteps),
              Diagnostic_Manager::getNumberOfMessagesReceived()/max(1.,numSteps));
#endif
    }

    delete tp;
  }
  
  
  if( np>1 )
  {
    // In parallel we print some timings for each processor
    for( int fileio=0; fileio<2; fileio++ )
    {
      FILE *output = fileio==0 ? logFile : file;
      fflush(output);
      fPrintF(output,"\n"
	      " ------- Summary: Timings per processor -----------\n"
	      "   p   ");
      for( int i=0; i<Parameters::maximumNumberOfTimings; i++ )
      { // output a short-form name (7 chars)
	if( timingName[i]!="" && maxTiming(i)!=0. )  
	{
          aString shortName="       ";
          int m=0;
	  for( int s=0; m<7 && s<timingName[i].length(); s++ ) 
	  { // strip off blanks
	    if( timingName[i][s]!=' ' ) {shortName[m]=timingName[i][s]; m++;} //
	  }
          fPrintF(output,"%7.7s ",(const char*)shortName);
	}
      }
      fPrintF(output,"\n");
      fflush(output);
      RealArray timingLocal(timing.dimension(0));
      for( int p=0; p<np; p++ )
      {
	// Note -- it did not work very well to have processor p try to write results, so instead
        // we copy results to processor 0 to print 
        timingLocal=timing;
        broadCast(timingLocal,p);  // send timing info from processor p   -- don't need a broad cast here **fix**
	fPrintF(output,"%4i : ",p);
	for( int i=0; i<Parameters::maximumNumberOfTimings; i++ )
	{
	  if( timingName[i]!="" && maxTiming(i)!=0. )    
	    fPrintF(output,"%7.1e ",timingLocal(i));
	}
	fflush(output);
	fPrintF(output,"\n");
      }
      fPrintF(output,"\n");
      fflush(output);
    }
  }
  
  printF("\n >>>> See the log file for further timings, memory usage and other statistics <<<< \n\n");

  // reset times
  timing(Parameters::timeForPlotting)+=timeForWaiting;
  timing(Parameters::totalTime)+=timeForWaiting;
  timing(Parameters::timeForAdvance)+=timeForWaiting;

  return 0;
}

// ===================================================================================================================
/// \brief Output information about the memory usage. 
/// \details This information is normally printed at the end of the run.
/// \param file (input) : output to this file.
// ===================================================================================================================
int DomainSolver::
printMemoryUsage(FILE *file /* = stdout */)
{
  const int np= max(1,Communication_Manager::numberOfProcessors());

  // These next functions may require communication
  //  const real totalArrayMemoryInUse= Diagnostic_Manager::getTotalArrayMemoryInUse();
  // const real totalMemoryInUse     = Diagnostic_Manager::getTotalMemoryInUse();

  real mem=Overture::getCurrentMemoryUsage();
  real maxMem=ParallelUtility::getMaxValue(mem);  // max over all processors
  real minMem=ParallelUtility::getMinValue(mem);  // min over all processors
  real totalMem=ParallelUtility::getSum(mem);     // sum of all processors
  real aveMem=totalMem/np;
  real maxMemRecorded=ParallelUtility::getMaxValue(Overture::getMaximumMemoryUsage());

  //  if( parameters.dbase.get<int >("myid")!=0 ) return 0;  // ** *wdh* 050409 not here -- display below

  const real megaByte=1024.*1024;

  // count the total number of grid points
  real totalNumberOfGridPoints=0.;
  int grid;
  for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    totalNumberOfGridPoints+=numberOfGridPoints[grid];
  
  int nSpace=30;
  aString dots="...................................................................";

  if( parameters.dbase.get<int >("myid")==0 )
    fprintf(file,"\n\n"
           "  --------------------------------------------------------- \n"
           "   Memory usage                   Mbytes    real/point   percent   \n");
  enum
  {
    memoryForCompositeGrid,
    memoryForOperators,
    memoryForInterpolant,
    memoryForCompositeGridSolver,
    memoryForGridFunctions,
    memoryForPressureSolver,
    memoryForImplicitSolver,
    memoryForWorkSpace,
    memoryForLineSolver,
    memoryTotal,
    numberOfMemoryItems
  };

  aString memoryName[numberOfMemoryItems]=
  {
    "CompositeGrid and u",
    "finite difference operators",
    "Interpolant",
    "CompositeGridSolver",
    "  grid functions",
    "  pressure solver",
    "  implicit solver",
    "  work space",
    "  line solver",
    "total (of above items)"
  };
  real memory[numberOfMemoryItems];
  memory[memoryTotal]=0.;
  memory[memoryForCompositeGrid]=gf[current].cg.sizeOf(); // +gf[current].u.sizeOf();
  memory[memoryTotal]+=memory[memoryForCompositeGrid];
  
  memory[memoryForOperators]=finiteDifferenceOperators.sizeOf();
  memory[memoryTotal]+=memory[memoryForOperators];

  if( cg.rcData->interpolant!=NULL )
  {
    memory[memoryForInterpolant]=cg.rcData->interpolant->sizeOf();
    memory[memoryTotal]+=memory[memoryForInterpolant];
  }

  memory[memoryForCompositeGridSolver]=sizeOf(file);
  memory[memoryTotal]+=memory[memoryForCompositeGridSolver];

  memory[memoryForGridFunctions]=0.;
  int i;
  for( i=0; i<DomainSolver::maximumNumberOfGridFunctionsToUse; i++ )
    memory[memoryForGridFunctions]+=gf[i].sizeOf()-
                                    gf[i].cg.sizeOf();   // do not count the cg;

  for( i=0; i<4; i++ )
    memory[memoryForGridFunctions]+=fn[i].sizeOf();

  memory[memoryForGridFunctions]+=pressureRightHandSide.sizeOf();
  if( pp!=NULL )
     memory[memoryForGridFunctions]+=p().sizeOf();  

  memory[memoryForPressureSolver]=poissonCoefficients.sizeOf();
  if( poisson!=NULL )
    memory[memoryForPressureSolver]+=poisson->sizeOf();
  
  memory[memoryForImplicitSolver]=0.;
  for( i=0; i<numberOfImplicitSolvers; i++ )
    memory[memoryForImplicitSolver]+=implicitSolver[i].sizeOf();
  memory[memoryForImplicitSolver]+=coeff.sizeOf();


  memory[memoryForWorkSpace]=0.;
//   if( pWorkSpace!=NULL )
//   {
//     real wSize=0.;
//     // only count unique work spaces
//     for( grid=0; grid<cg.numberOfComponentGrids(); grid++)
//     {
//       if( grid==0 || min(abs( workSpaceIndex(grid)-workSpaceIndex(Range(0,grid-1)))) !=0 )
// 	wSize+=workSpace(grid).sizeOf();
//     }
//     memory[memoryForWorkSpace]=wSize;
//   }
 

  memory[memoryForLineSolver]=0.;
  if( pLineSolve!=NULL ) 
     memory[memoryForLineSolver]=pLineSolve->sizeOf();
  
  if( parameters.dbase.get<int >("myid")==0 )
  {
    for( i=0; i<numberOfMemoryItems; i++ )
    {
      fprintf(file," %s%s%9.2f  %9.2f     %5.1f  \n",
	      (const char*)memoryName[i],(const char*)dots(0,max(0,nSpace-memoryName[i].length())),
	      memory[i]/megaByte,memory[i]/sizeof(real)/totalNumberOfGridPoints, 100.*memory[i]/memory[memoryTotal]);
    }
    fprintf(file,"\n **Bytes per grid point = %9.3e/%i = %9.3e\n",
	    memory[memoryTotal],totalNumberOfGridPoints,memory[memoryTotal]/totalNumberOfGridPoints);
    fprintf(file," **number of reals per grid point = %9.3e\n\n",memory[memoryTotal]/sizeof(real)/totalNumberOfGridPoints);
  }
  
  if( parameters.dbase.get<int >("debug") & 2 )
    finiteDifferenceOperators.sizeOf(file);

  if( true || parameters.dbase.get<int >("debug") & 2 )
  {
    if( parameters.dbase.get<int >("myid")==0 ) fprintf(file,"\n +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
      cg[grid].displayComputedGeometry(file);
    if( parameters.dbase.get<int >("myid")==0 ) fprintf(file," +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
  }
  if( parameters.dbase.get<int >("myid")==0 )
  {
    fprintf(file,"===== memory per-proc: [min=%g,ave=%g,max=%g](Mb), max-recorded=%g (Mb), total=%g (Mb)\n",
	    minMem,aveMem,maxMem,maxMemRecorded,totalMem);

//     fprintf(file,"************************************************************************* \n"
// 	    " Here is what memory is used by A++ arrays \n"
// 	    " total array memory in use = %10.3fM, (%10.3fM including overhead)\n"
// 	    " (This info is obtained by running cg with the `memory' option)\n"
// 	    "*************************************************************************\n\n",
// 	    totalArrayMemoryInUse/megaByte,totalMemoryInUse/megaByte);
    
  }
  
  return 0;
}

