#include "DomainSolver.h"
#include "GenericGraphicsInterface.h"
#include "ParallelUtility.h"
#include "ExternalBoundaryData.h"

namespace
{
enum UserDefinedBoundaryConditions
{
  variableInflow,
  timeDependentInflow,
  perturbedShearFlow,
  wallWithScalarFlux,
  axisymmetricRotation,
  linearRampInX,
  linearRampInY,
  ablProfile,
  boundaryDataFromAFile // reads probe bounding box data for e.g.
};
 
}

// =========================================================================================
/// \brief Interactively define user specific values for boundary conditions. 
/// \details This function will be called when interactively choosing boundary conditions and the
/// option userDefinedBoundaryData is used.
///
/// You may add a time dependence to an existing boundary condition or you may define a 
/// new boundary condition. In this function you should prompt for the boundary condition
/// to be used as well as any parameters that will be needed. Parameters can be saved using
/// the setUserBoundaryConditionParameters function. 
///
/// \note:  The actual boundary values are NOT assigned in this routine. This is done in the
///    userDefinedBoundaryValues function.
/// 
/// \param side,axis,grid (input): assign boundary conditions for this face.
///
// =========================================================================================
int Parameters::
chooseUserDefinedBoundaryValues(int side, int axis, int grid, CompositeGrid & cg)
{
  Parameters & parameters = *this;
  assert( parameters.dbase.get<GenericGraphicsInterface* >("ps") !=NULL );
  GenericGraphicsInterface & gi = *parameters.dbase.get<GenericGraphicsInterface* >("ps");


  printf("Choose boundary values for (side,axis,grid)=(%i,%i,%i) gridName=%s\n",side,axis,grid,
	 (const char*)cg[grid].getName());

  aString menu[]=
  {
    "!user boundary values",
    "variable inflow",
    "time dependent inflow",
    "perturbed shear flow",
    "wall with scalar flux",
    "axisymmetric rotation",
    "linear ramp in x",
    "linear ramp in y",
    "abl profile",
    "boundary data from a file",
    "done",
    ""
  };
  aString answer,answer2;
  Index Ib1,Ib2,Ib3;
//   const int pc=parameters.dbase.get<int >("pc");
//   const int uc=parameters.dbase.get<int >("uc");
//   const int vc=parameters.dbase.get<int >("vc");
//   const int wc=parameters.dbase.get<int >("wc");
  
  for(;;)
  {
    gi.getMenuItem(menu,answer,"choose a menu item");
    if( answer=="done" || answer=="exit" )
    {
      break;
    }
    else if( answer=="variable inflow" )
    {
      parameters.setUserBcType(side,axis,grid,variableInflow);    // set the bcType to be a unique value.
      parameters.setBcIsTimeDependent(side,axis,grid,false);  // this condition is NOT time dependent

      gi.inputString(answer2,"Enter u,v,w");
      real u0=1., v0=0., w0=0.;
      if( answer2!="" )
      {
	sScanF(answer2,"%e %e %e",&u0,&v0,&w0);
      }
      printf("***userDefinedBoundaryValues: assign variable inflow on (side,axis,grid)=(%i,%i,%i)\n",
	     side,axis,grid);

      RealArray values(3);
      values(0)=u0;
      values(1)=v0;
      values(2)=w0;
      // save the parameters to be used when evaluating the time dependent BC's:
      parameters.setUserBoundaryConditionParameters(side,axis,grid,values);

    }
    else if( answer=="time dependent inflow" )
    {
      parameters.setUserBcType(side,axis,grid,timeDependentInflow);  // set the bcType to be a unique value.
      parameters.setBcIsTimeDependent(side,axis,grid,true);      // this condition is time dependent

      gi.inputString(answer2,"Enter u,v,w");
      real u0=1., v0=0., w0=0.;
      if( answer2!="" )
      {
	sScanF(answer2,"%e %e %e",&u0,&v0,&w0);
      }
      printf("***userDefinedBoundaryValues: assign time dependent inflow on (side,axis,grid)=(%i,%i,%i)\n",
	     side,axis,grid);

      RealArray values(3);
      values(0)=u0;
      values(1)=v0;
      values(2)=w0;
      // save the parameters to be used when evaluating the time dependent BC's:
      parameters.setUserBoundaryConditionParameters(side,axis,grid,values);

    }    
    else if( answer=="perturbed shear flow" )
    {
      parameters.setUserBcType(side,axis,grid,perturbedShearFlow);  // set the bcType to be a unique value.

      gi.inputString(answer2,"Enter u1,u2,beta,y0,f0,amp");
      real u1=2., u2=1.,beta=50., y0=.5, f0=1., amp=0.;
      if( answer2!="" )
      {
	sScanF(answer2,"%e %e %e %e %e %e",&u1,&u2,&beta,&y0,&f0,&amp);
      }
      printf("***userDefinedBoundaryValues: assign perturbed shear flow on (side,axis,grid)=(%i,%i,%i)\n",
	     side,axis,grid);

      RealArray values(6);
      values(0)=u1;
      values(1)=u2;
      values(2)=beta;
      values(3)=y0;
      values(4)=f0;
      values(5)=amp;
      // save the parameters to be used when evaluating the time dependent BC's:
      parameters.setUserBoundaryConditionParameters(side,axis,grid,values);

      parameters.setBcIsTimeDependent(side,axis,grid,amp!=0.);      // is this condition is time dependent?

    }
    else if( answer=="wall with scalar flux" )
    {
      parameters.setUserBcType(side,axis,grid,wallWithScalarFlux);  // set the bcType to be a unique value.

      gi.inputString(answer2,"Enter amplitude,radius and position: amp,radius,x0,y0,z0");
      real amp=1.,radius=1.,x0=0.,y0=0.,z0=0;
      if( answer2!="" )
      {
	sScanF(answer2,"%e %e %e %e %e %e",&amp,&radius,&x0,&y0,&z0);
      }
      printf("***userDefinedBoundaryValues: wall with scalar flux on (side,axis,grid)=(%i,%i,%i)\n"
             "                              amp=%8.2e, radius=%8.2e, x=(%8.2e,%8.2e,%8.2e)\n",
	     side,axis,grid,amp,radius,x0,y0,z0);

      RealArray values(5);
      values(0)=amp;
      values(1)=radius;
      values(2)=x0;
      values(3)=y0;
      values(4)=z0;
      // save the parameters to be used when evaluating the time dependent BC's:
      parameters.setUserBoundaryConditionParameters(side,axis,grid,values);

      parameters.setBcIsTimeDependent(side,axis,grid,false);    // is this condition is time dependent?

    }
    else if( answer=="axisymmetric rotation" )
    {
      parameters.setUserBcType(side,axis,grid,axisymmetricRotation);    // set the bcType to be a unique value.
      parameters.setBcIsTimeDependent(side,axis,grid,true);  // this condition is time dependent

      gi.inputString(answer2,"Enter d(omega)/dt, u, v, t, ramp time");
      real odot=0., u0=1., v0=0., te0=1., trmp=-1;
      if( answer2!="" )
      {
	sScanF(answer2,"%e %e %e %e %e",&odot,&u0,&v0,&te0, &trmp);
      }
      printf("***userDefinedBoundaryValues: assign axisymmetric rotation on (side,axis,grid)=(%i,%i,%i)\n",
	     side,axis,grid);

      RealArray values(5);
      values(0)=odot;
      values(1)=u0;
      values(2)=v0;
      values(3)=te0;
      values(4)=trmp;
      // save the parameters to be used when evaluating the time dependent BC's:
      parameters.setUserBoundaryConditionParameters(side,axis,grid,values);
      if ( te0<=0. && false) 
	{
	  parameters.dbase.get<RealArray>("bcData")(parameters.dbase.get<int >("numberOfComponents")+1,side,axis,grid) = 1;
	  cout<<"adiabatic wall on "<<side<<"  "<<axis<<endl;
	}

    }    
    else if( answer=="linear ramp in x" )
    {
      parameters.setUserBcType(side,axis,grid,linearRampInX);    // set the bcType to be a unique value.
      parameters.setBcIsTimeDependent(side,axis,grid,true);  // this condition is time dependent

      RealArray values(dbase.get<int>("numberOfComponents")*3);
      values = -1;
      for( ;; )
	{
	  gi.inputString(answer2,"Enter component, f0 (constant), slope, ramp time");
	  int comp=-1;
	  real f0=0, slope=0, trmp=0;
	  if ( answer2=="done")
	    break;
	  else if( answer2!="" )
	    {
	      sScanF(answer2,"%i %e %e %e",&comp, &f0,&slope, &trmp);
	    }
	  
	  if ( comp>-1 )
	    {
	      values(3*comp+0)=f0;
	      values(3*comp+1)=slope;
	      values(3*comp+2)=trmp;
	    }
	  else
	    break;
	}
      // save the parameters to be used when evaluating the time dependent BC's:
      parameters.setUserBoundaryConditionParameters(side,axis,grid,values);
      
      printf("***userDefinedBoundaryValues: assign axisymmetric rotation on (side,axis,grid)=(%i,%i,%i)\n",
	     side,axis,grid);
    }    
    else if( answer=="linear ramp in y" )
    {
      parameters.setUserBcType(side,axis,grid,linearRampInY);    // set the bcType to be a unique value.
      parameters.setBcIsTimeDependent(side,axis,grid,true);  // this condition is time dependent

      RealArray values(dbase.get<int>("numberOfComponents")*3);
      values = -1;
      for( ;; )
	{
	  gi.inputString(answer2,"Enter component, f0 (constant), slope, ramp time");
	  int comp=-1;
	  real f0=0, slope=0, trmp=0;
	  if ( answer2=="done")
	    break;
	  else if( answer2!="" )
	    {
	      sScanF(answer2,"%i %e %e %e",&comp, &f0,&slope, &trmp);
	    }
	  
	  if ( comp>-1 )
	    {
	      values(3*comp+0)=f0;
	      values(3*comp+1)=slope;
	      values(3*comp+2)=trmp;
	    }
	  else
	    break;
	}
      // save the parameters to be used when evaluating the time dependent BC's:
      parameters.setUserBoundaryConditionParameters(side,axis,grid,values);
      
      printf("***userDefinedBoundaryValues: assign axisymmetric rotation on (side,axis,grid)=(%i,%i,%i)\n",
	     side,axis,grid);
    }    
    else if( answer=="abl profile" )
      {
	parameters.setUserBcType(side,axis,grid,ablProfile);
	parameters.setBcIsTimeDependent(side,axis,grid,false);

	gi.inputString(answer2,"Enter u_ref, z_ref, alpha");
	real u_ref=1., z_ref=1., alpha=1.;
	if( answer2!="" )
	  {
	    sScanF(answer2,"%e %e %e",&u_ref,&z_ref,&alpha);
	  }
	printf("***userDefinedBoundaryValues: assign abl profile on (side,axis,grid)=(%i,%i,%i)\n",
	       side,axis,grid);

	RealArray values(3);
	values(0) = u_ref;
	values(1) = z_ref;
	values(2) = alpha;

	// save parameters to be used when evaluating the time dependent BC's:
	parameters.setUserBoundaryConditionParameters(side,axis,grid,values);
      }
    else if( answer=="boundary data from a file" )
    {
      parameters.setUserBcType(side,axis,grid,boundaryDataFromAFile);    // set the bcType to be a unique value.
      parameters.setBcIsTimeDependent(side,axis,grid,true);  // this condition is time dependent

      // The ExternalBoundaryData class knows how to deal with external data formats
      // that hold time depedent boundary data.
      if( !parameters.dbase.has_key("externalBoundaryData") ) 
      {
        parameters.dbase.put<ExternalBoundaryData*>("externalBoundaryData");
        parameters.dbase.get<ExternalBoundaryData*>("externalBoundaryData")=new ExternalBoundaryData;
        
      }
      parameters.dbase.get<ExternalBoundaryData*>("externalBoundaryData")->update(gi);


//       aString probeFileName;
//       gi.inputString(probeFileName,"Enter the name of the probe bounding box data file");
//       printF("probeFileName=[%s]\n",(const char*)probeFileName);
      
//       assert( !parameters.dbase.has_key("boundaryDataFile") );
      

//       if( !parameters.dbase.has_key("boundaryDataFile") ) 
//       {
//         parameters.dbase.put<GenericDataBase*>("boundaryDataFile");
//         parameters.dbase.get<GenericDataBase*>("boundaryDataFile")=NULL;
//       }
//       GenericDataBase *& pdb = parameters.dbase.get<GenericDataBase*>("boundaryDataFile");
      
//       if( pdb==NULL )
//       {
// 	pdb = new HDF_DataBase;  // ************ who will delete ?? ***********************

//         printF("userDefinedBoundaryValues: opening the data base file %s for the bounding box probe info.\n",
// 	       (const char*)probeFileName);

//         HDF_DataBase & db = ( HDF_DataBase &)(*pdb);
//         db.setMode(GenericDataBase::noStreamMode);
// 	db.mount(probeFileName,"R");    // open the data base, R=read-only
//       }
//       assert( pdb!=NULL );
//       HDF_DataBase & db = ( HDF_DataBase &)(*pdb);

//       int numberOfTimes=-1;
//       db.get(numberOfTimes,"numberOfTimes");
//       RealArray times;
//       db.get(times,"times");

//       printF(" Found probe data: numberOfTimes=%i\n",numberOfTimes);
//       ::display(times,"The solution was saved at these times","%6.3f ");

//       OV_ABORT("finish me");


    }
    else
    {
      printF("Unknown answer =[%s]\n",(const char*)answer2);
      gi.stopReadingCommandFile();
    }
  }

  return 0;
}




// =========================================================================================
/// \brief Assign user specific values for boundary conditions. 
/// \details The user may fill in
/// the boundaryData array with right-hand-side values for boundary conditions.
/// The user is also required to provide the time derivative of the boundary values.
/// 
/// \note  This function shows examples of applying a user defined boundary condition as
/// well as adding a time dependence to some pre-defined boundary conditions.
///
/// \param t (input) : current time.
/// \param u (input) : the current solution.
/// \param gridVelocity (input) : if the grid is moving, this is the grid velocity.
/// \param grid (input): the component grid we are assigning.
/// \param forcingType (input) : if forcingType==computeForcing then return the rhs for the 
///  boundary condition; if forcingType==computeTimeDerivativeOfForcing then return the 
///   first time derivative of the forcing.
///
// =========================================================================================
int DomainSolver::
userDefinedBoundaryValues(const real & t, 
			  realMappedGridFunction & u, 
			  realMappedGridFunction & gridVelocity,
			  const int & grid,
			  int side0 /* = -1 */,
			  int axis0 /* = -1 */,
			  ForcingTypeEnum forcingType /* =computeForcing */)
{
  // printf("***userDefinedBoundaryValues\n");

  MappedGrid & mg = *u.getMappedGrid();

  assert( side0>=-1 && side0<2 );
  assert( axis0>=-1 && axis0<parameters.dbase.get<int >("numberOfDimensions") );
  
  const int axisStart= axis0==-1 ? 0 : axis0;
  const int axisEnd  = axis0==-1 ? parameters.dbase.get<int >("numberOfDimensions")-1 : axis0;
  const int sideStart= side0==-1 ? 0 : side0;
  const int sideEnd  = side0==-1 ? 1 : side0;

  int numberOfSidesAssigned=0;

  Range C(0,parameters.dbase.get<int >("numberOfComponents")-1);
  const int pc=parameters.dbase.get<int >("pc");
  const int uc=parameters.dbase.get<int >("uc");
  const int vc=parameters.dbase.get<int >("vc");
  const int wc=parameters.dbase.get<int >("wc");
  const int tc=parameters.dbase.get<int >("tc");
  
  const bool gridIsMoving = parameters.gridIsMoving(grid); // true if the grid is moving

  // uncomment the next two lines if you want the grid points.
  // c.update(MappedGrid::THEvertex);
  // const realArray & vertex = c.vertex();  // here is the array of grid points
  
 int includeGhost=1;
  #ifdef USE_PPP
    realSerialArray uLocal; getLocalArrayWithGhostBoundaries(u,uLocal);
  #else
    const realSerialArray & uLocal = u;
  #endif

  const int numberOfDimensions = mg.numberOfDimensions();
  
  int axis;
  Index Ib1,Ib2,Ib3;
  
  for( axis=axisStart; axis<=axisEnd; axis++ )
  {
    for( int side=sideStart; side<=sideEnd; side++ )
    {
      if( parameters.userBcType(side,axis,grid)==variableInflow )
      {
        RealArray values(3);
	parameters.getUserBoundaryConditionParameters(side,axis,grid,values);
        // printf("***userDefinedBoundaryValues: assign (side,axis,grid)=(%i,%i,%i) variableInflow force=%i, "
        //        "values=%f,%f,%f\n",side,axis,grid,(int)forcingType,values(0),values(1),values(2));

        getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
	
	// kkc 070130 BILL : this section of code does nothing if mg.bC != inflowWithVelocityGiven, so right now
	//                   I assume this is only called in that instance obviating the need for this check
	//        if( mg.boundaryCondition(side,axis)==Parameters::inflowWithVelocityGiven )
	//	{
	// assign values for inflow velocity: The first 3 components are the values of (u,v,w) on the boundary.
	numberOfSidesAssigned++;

	bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	if( !ok ) continue;  // no points on this processor

	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
	real u0=values(0);
	real v0=values(1);
	real w0=values(2);
	// just set to constant values for now.
	bd(Ib1,Ib2,Ib3,uc)=u0;
	bd(Ib1,Ib2,Ib3,vc)=v0;
	if( numberOfDimensions>2 )
	  bd(Ib1,Ib2,Ib3,wc)=w0;
      }
      else if( parameters.userBcType(side,axis,grid)==timeDependentInflow )
      {

        RealArray values(3);
	parameters.getUserBoundaryConditionParameters(side,axis,grid,values);

        // printf("***userDefinedBoundaryValues: assign (side,axis,grid)=(%i,%i,%i) timeDependentInflow force=%i, "
        //        "values=%f,%f,%f\n",side,axis,grid,(int)forcingType,values(0),values(1),values(2));

        getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
	
	bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	if( !ok ) continue;  // no points on this processor

	// kkc 070130 BILL : this section of code does nothing if mg.bC != inflowWithVelocityGiven, so right now
	//                   I assume this is only called in that instance obviating the need for this check

	//        if( mg.boundaryCondition(side,axis)==Parameters::inflowWithVelocityGiven )
	//	{
	// printf("***userDefinedBoundaryValues: assign (side,axis,grid)=(%i,%i,%i) timeDependentInflow force=%i,"
	//	 " values=%f,%f,%f\n",side,axis,grid,(int)forcingType,values(0),values(1),values(2));

	// assign values for inflow velocity: The first 3 components are the values of (u,v,w) on the boundary.
	numberOfSidesAssigned++;

	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);

	real u0=values(0);
	real v0=values(1);
	real w0=values(2);

	if( parameters.bcType(side,axis,grid)==Parameters::parabolicInflowUserDefinedTimeDependence )
	{
	  // give time dependence to to a parabolic inflow profile.
	  real factor;
	  if( forcingType==computeForcing )
	    factor=min(1.,t*2.);  // ramp values for u in time.
	  else 
	    factor=t<.5 ? 2.*u0 : 0.;  // time derivative of the forcing
  
	  // the parabolic profile is saved in the ghost line of the "bd" array.
	  Index Ig1,Ig2,Ig3;
	  getGhostIndex(mg.gridIndexRange(),side,axis,Ig1,Ig2,Ig3);
  	  ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ig1,Ig2,Ig3,includeGhost);

	  Range C(uc,uc+numberOfDimensions-1);
	  bd(Ib1,Ib2,Ib3,C)= factor*bd(Ig1,Ig2,Ig3,C);
	    
	}
	else
	{
	  if( forcingType==computeForcing )
	    bd(Ib1,Ib2,Ib3,uc)=u0*min(1.,t*2.);  // ramp values for u in time.
	  else 
	    bd(Ib1,Ib2,Ib3,uc)=t<.5 ? 2.*u0 : 0.;  // time derivative of the forcing
  
	  bd(Ib1,Ib2,Ib3,vc)=v0;
	  if( numberOfDimensions>2 )
	    bd(Ib1,Ib2,Ib3,wc)=w0;
	}
	  
	//	}
      }
      else if( parameters.userBcType(side,axis,grid)==perturbedShearFlow )
      {

	// kkc 070130 BILL : this section of code does nothing if mg.bC != inflowWithVelocityGiven, so right now
	//                   I assume this is only called in that instance obviating the need for this check
	//        if( mg.boundaryCondition(side,axis)==Parameters::inflowWithVelocityGiven )
	//	{

	// printf("***userDefinedBoundaryValues: assign (side,axis,grid)=(%i,%i,%i) timeDependentInflow force=%i, "
	//        "values=%f,%f,%f\n",side,axis,grid,(int)forcingType,values(0),values(1),values(2));

	getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
	
	// printf("***userDefinedBoundaryValues: assign (side,axis,grid)=(%i,%i,%i) timeDependentInflow force=%i,"
	//	 " values=%f,%f,%f\n",side,axis,grid,(int)forcingType,values(0),values(1),values(2));

	numberOfSidesAssigned++;
	bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	if( !ok ) continue;  // no points on this processor

	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);

	RealArray values(6);
	parameters.getUserBoundaryConditionParameters(side,axis,grid,values);
	real u1  =values(0);
	real u2  =values(1);
	real beta=values(2);
	real y0  =values(3);
	real f0  =values(4);
	real amp =values(5);

	if( t==0. )
	  printf(" ASSIGN perturbed shear flow: u1=%8.2e u2=%8.2e beta=%8.2e y0=%8.2e f0=%8.2e amp=%8.2e\n",
		 u1,u2,beta,y0,f0,amp);
	  
	  
	Range C(uc,uc+numberOfDimensions-1);
	bd(Ib1,Ib2,Ib3,C)=0.;
#define STIME(t) ( 1.+ amp*( .5*sin(2.*Pi*f0*(t))+(1./3.)*sin(3.*Pi*f0*(t))+.25*sin(4.*Pi*f0*(t))+\
                     .2*sin(5.*Pi*f0*(t))) )
#define STIME_T(t) ( amp*Pi*f0*( cos(2.*Pi*f0*(t))+cos(3.*Pi*f0*(t))+cos(4.*Pi*f0*(t))+cos(5.*Pi*f0*(t))) )

#define SHEAR(y,t) ((u1+u2)*.5+(u1-u2)*.5*tanh(beta*((y)-y0)))*STIME(t)
#define SHEAR_T(y,t) ((u1+u2)*.5+(u1-u2)*.5*tanh(beta*((y)-y0)))*STIME_T(t)
	if( mg.isRectangular() )
	{
	  real dx[3],xab[2][3];
	  mg.getRectangularGridParameters( dx, xab );
	  const int i0a=mg.gridIndexRange(0,0);
	  const int i1a=mg.gridIndexRange(0,1);
	  const int i2a=mg.gridIndexRange(0,2);
                
	  const real xa=xab[0][0], dx0=dx[0];
	  const real ya=xab[0][1], dy0=dx[1];
	  const real za=xab[0][2], dz0=dx[2];

#define X0(i0,i1,i2) (xa+dx0*(i0-i0a))
#define X1(i0,i1,i2) (ya+dy0*(i1-i1a))
#define X2(i0,i1,i2) (za+dz0*(i2-i2a))

	  RealArray y(1,Ib2);
	  y(0,Ib2).seqAdd(X1(i0a,Ib2.getBase(),i2a),dy0);
	  if( forcingType==computeForcing )
	    bd(Ib1,Ib2,Ib3,uc)=SHEAR(y,t);
	  else
	    bd(Ib1,Ib2,Ib3,uc)=SHEAR_T(y,t);
	}
	else
	{
	  mg.update(MappedGrid::THEvertex);
	  const RealArray & vertex = mg.vertex().getLocalArray(); // no need to use if rectangular ************
	  
	  if( forcingType==computeForcing )
	    bd(Ib1,Ib2,Ib3,uc)=SHEAR(vertex(Ib1,Ib2,Ib3,axis2),t);
	  else
	    bd(Ib1,Ib2,Ib3,uc)=SHEAR_T(vertex(Ib1,Ib2,Ib3,axis2),t);

	}
	//	}
      }
      else if( parameters.userBcType(side,axis,grid)==wallWithScalarFlux )
      {
        if( mg.boundaryCondition(side,axis)==Parameters::noSlipWall ||
            mg.boundaryCondition(side,axis)==Parameters::slipWall )
	{

	  getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);
	
	  // printf("***userDefinedBoundaryValues: assign (side,axis,grid)=(%i,%i,%i) timeDependentInflow force=%i,"
	  //	 " values=%f,%f,%f\n",side,axis,grid,(int)forcingType,values(0),values(1),values(2));

	  numberOfSidesAssigned++;

	  bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	  if( !ok ) continue;  // no points on this processor

          RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);

	  RealArray values(5);
	  parameters.getUserBoundaryConditionParameters(side,axis,grid,values);
	  real amp=    values(0);
	  real radius= values(1);
	  real x0=     values(2);
	  real y0=     values(3);
	  real z0=     values(4);

          if( true )
	    printf(" ASSIGN wallWithScalarFlux: amp=%8.2e, radius=%8.2e, x=(%8.2e,%8.2e,%8.2e)\n",
		   amp,radius,x0,y0,z0);
	  
	  Range V(uc,uc+numberOfDimensions-1);
	  bd(Ib1,Ib2,Ib3,V)=0.;  // set values for the velocity components on the wall

          // Now set the RHS for the scalar BC:
          //     s.n = rhs
	  const int sc=parameters.dbase.get<int >("sc");
          assert( sc>=0 );

	  const RealArray & vertex = mg.vertex().getLocalArray();
          RealArray rad(Ib1,Ib2,Ib3);
          if( mg.numberOfDimensions()==2 )
  	    rad = SQR(vertex(Ib1,Ib2,Ib3,0)-x0)+SQR(vertex(Ib1,Ib2,Ib3,1)-y0);
          else
  	    rad = SQR(vertex(Ib1,Ib2,Ib3,0)-x0)+SQR(vertex(Ib1,Ib2,Ib3,1)-y0)+SQR(vertex(Ib1,Ib2,Ib3,2)-z0);
	  
          // note: rhs for neumann BC is filled into the bd array for points on the boundary
          bd(Ib1,Ib2,Ib3,sc)=0.;
          where( rad<SQR(radius) )
	  {
	    bd(Ib1,Ib2,Ib3,sc)=amp;
	  }
	  
	}

      }
      else if( parameters.userBcType(side,axis,grid)==axisymmetricRotation )
      {
        RealArray values(5);
	parameters.getUserBoundaryConditionParameters(side,axis,grid,values);
	printf("***userDefinedBoundaryValues: assign (side,axis,grid)=(%i,%i,%i) axisymmetric rotation =%i, "
	       "values=%f,%f,%f,%f,%f\n",side,axis,grid,(int)forcingType,values(0),values(1),values(2),values(3),values(4));

        const int extra=1;  // include ghost
        getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3,extra);
	//        getBoundaryIndex(mg.extendedIndexRange(),side,axis,Ib1,Ib2,Ib3,extra);

	if ( mg.isPeriodic((axis+1)%2) )
	{
	  if ( axis==1 )
	    Ib1 = Index(Ib1.getBase(), Ib1.getLength()-1);
	  else
	    Ib2 = Index(Ib2.getBase(), Ib2.getLength()-1);
	}
	// assign values for inflow velocity: The first 3 components are the values of (u,v,w) on the boundary.
	numberOfSidesAssigned++;

	bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	if( !ok ) continue;  // no points on this processor

	mg.update(MappedGrid::THEvertex);
        const RealArray & vertex = mg.vertex().getLocalArray();

	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);

	real odot=values(0);
	real u0=values(1);
	real v0=values(2);
	real te0=values(3);
	real trmp=values(4);

	// just set to constant values for now.
	real f = trmp>REAL_MIN ? min(1.,t/trmp) : 1;
	f = f*f*f;
	bd(Ib1,Ib2,Ib3,uc)=f*u0;
	bd(Ib1,Ib2,Ib3,vc)=f*v0;
	if( wc>=0 )
	{
	  bd(Ib1,Ib2,Ib3,wc)= f*odot*vertex(Ib1,Ib2,Ib3,1); // XXX assumed radial coordinate is 1!!!
	}

	if( tc>=0 ) 
	{
          // printf(" ********** axisymmetricRotation BC: set T = %8.2e\n",te0);
	  if ( te0>0. )
	  {
	    bd(Ib1,Ib2,Ib3,tc)=f*te0 + (1-f)*uLocal(Ib1,Ib2,Ib3,tc);
	  }
	  else
	  {
	    bd(Ib1,Ib2,Ib3,tc)= uLocal(Ib1,Ib2,Ib3,tc);
	  }
	}
      }
      else if( parameters.userBcType(side,axis,grid)==linearRampInX )
      {
        RealArray values(3*parameters.dbase.get<int >("numberOfComponents"));
	parameters.getUserBoundaryConditionParameters(side,axis,grid,values);
        // printf("***userDefinedBoundaryValues: assign (side,axis,grid)=(%i,%i,%i) variableInflow force=%i, "
        //        "values=%f,%f,%f\n",side,axis,grid,(int)forcingType,values(0),values(1),values(2));

        const int extra=1;  // include ghost
        getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3,extra);
	if ( mg.isPeriodic((axis+1)%2) )
	{
	  if ( axis==1 )
	    Ib1 = Index(Ib1.getBase(), Ib1.getLength()-1);
	  else
	    Ib2 = Index(Ib2.getBase(), Ib2.getLength()-1);
	}
	// assign values for inflow velocity: The first 3 components are the values of (u,v,w) on the boundary.
	numberOfSidesAssigned++;

	bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	if( !ok ) continue;  // no points on this processor


	mg.update(MappedGrid::THEvertex);
        const RealArray & vertex = mg.vertex().getLocalArray();

	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
	for ( int c=0; c<parameters.dbase.get<int >("numberOfComponents"); c++ )
	{
	  real f0 = values(3*c+0);
	  real slope=values(3*c+1);
	  real trmp = values(3*c+2);

	  // just set to constant values for now.
	  if ( trmp>=0. )
	  {
	    real f = trmp>REAL_MIN ? min(1.,t/trmp) : 1;
	    f = f*f*f;
		
	    bd(Ib1,Ib2,Ib3,c) = f*(vertex(Ib1,Ib2,Ib3,0)*slope + f0) + (1-f)*uLocal(Ib1,Ib2,Ib3,c);
	  }
	  else
	    bd(Ib1,Ib2,Ib3,c) = uLocal(Ib1,Ib2,Ib3,c);

	}
      }
      else if( parameters.userBcType(side,axis,grid)==linearRampInY )
      {
        RealArray values(3*parameters.dbase.get<int >("numberOfComponents"));
	parameters.getUserBoundaryConditionParameters(side,axis,grid,values);
        // printf("***userDefinedBoundaryValues: assign (side,axis,grid)=(%i,%i,%i) variableInflow force=%i, "
        //        "values=%f,%f,%f\n",side,axis,grid,(int)forcingType,values(0),values(1),values(2));

        const int extra=1;  // include ghost
        getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3,extra);
	if ( mg.isPeriodic((axis+1)%2) )
	{
	  if ( axis==1 )
	    Ib1 = Index(Ib1.getBase(), Ib1.getLength()-1);
	  else
	    Ib2 = Index(Ib2.getBase(), Ib2.getLength()-1);
	}
	// assign values for inflow velocity: The first 3 components are the values of (u,v,w) on the boundary.
	numberOfSidesAssigned++;

	bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	if( !ok ) continue;  // no points on this processor


	mg.update(MappedGrid::THEvertex);
        const RealArray & vertex = mg.vertex().getLocalArray();

	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
	for ( int c=0; c<parameters.dbase.get<int >("numberOfComponents"); c++ )
	{
	  real f0 = values(3*c+0);
	  real slope=values(3*c+1);
	  real trmp = values(3*c+2);

	  // just set to constant values for now.
	  if ( trmp>=0. )
	  {
	    real f = trmp>REAL_MIN ? min(1.,t/trmp) : 1;
	    f = f*f*f;
		
	    bd(Ib1,Ib2,Ib3,c) = f*(vertex(Ib1,Ib2,Ib3,1)*slope + f0) + (1-f)*uLocal(Ib1,Ib2,Ib3,c);
	  }
	  else
	    bd(Ib1,Ib2,Ib3,c) = uLocal(Ib1,Ib2,Ib3,c);

	}
      }
      else if(parameters.userBcType(side,axis,grid)==ablProfile )
	{

	  RealArray values(3);
	  parameters.getUserBoundaryConditionParameters(side,axis,grid,values);

	  real u_ref, z_ref, alpha;
	  u_ref = values(0);
	  z_ref = values(1);
	  alpha = values(2);
	  printf("***userDefinedBoundaryValues: assign (side,axis,grid)=(%i,%i,%i) abl profile = %f %f %f\n",
		 side,axis,grid,u_ref,z_ref,alpha);
	  
	  getBoundaryIndex(mg.gridIndexRange(),side,axis,Ib1,Ib2,Ib3);

	  numberOfSidesAssigned++;

	  bool ok=ParallelUtility::getLocalArrayBounds(u,uLocal,Ib1,Ib2,Ib3,includeGhost);
	  if( !ok ) continue;  // no points on this processor

	  RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);

	  Range C(uc,uc+numberOfDimensions-1);
	  bd(Ib1,Ib2,Ib3,C) = 0.;

#define X0(i0,i1,i2) (xa+dx0*(i0-i0a))
#define X1(i0,i1,i2) (ya+dy0*(i1-i1a))
#define X2(i0,i1,i2) (za+dz0*(i2-i2a))

	  if( mg.isRectangular() )
	    {
	      real dx[3], xab[2][3];
	      mg.getRectangularGridParameters(dx,xab);
	      const int i0a = mg.gridIndexRange(0,0);
	      const int i1a = mg.gridIndexRange(0,1);
	      const int i2a = mg.gridIndexRange(0,2);

	      const real xa = xab[0][0], dx0 = dx[0];
	      const real ya = xab[0][1], dy0 = dx[1];
	      const real za = xab[0][2], dz0 = dx[2];

	      for ( int ib3=Ib3.getBase(); ib3<=Ib3.getBound(); ib3++ )
		for ( int ib2=Ib2.getBase(); ib2<=Ib2.getBound(); ib2++ )
		  for ( int ib1=Ib1.getBase(); ib1<=Ib1.getBound(); ib1++ )
		    bd(ib1,ib2,ib3,uc) = u_ref * pow(X1(ib1,ib2,ib3)/z_ref,alpha);
	    }
	  else
	    {
	      mg.update(MappedGrid::THEvertex);
	      const RealArray & vertex = mg.vertex().getLocalArray();
	      bd(Ib1,Ib2,Ib3,uc) = u_ref * pow(abs(vertex(Ib1,Ib2,Ib3,1))/z_ref,alpha);
	    }

	}
      else if( parameters.userBcType(side,axis,grid)==boundaryDataFromAFile )
      {
        ExternalBoundaryData & ebd = *parameters.dbase.get<ExternalBoundaryData*>("externalBoundaryData");
	RealArray & bd = parameters.getBoundaryData(side,axis,grid,mg);
        ebd.getBoundaryData( t, cg, side, axis, grid, bd );
	
      }
      else
      {
	// printf(" userDefinedBoundaryValues: mg.boundaryCondition(side,axis)=%i \n",mg.boundaryCondition(side,axis));
	
      }
    }
    
  }
  
  return numberOfSidesAssigned;
}
